"""
Module `chatette.units.slot`
Cotains the definition and rule contents characterizing slots (i.e. entities).
"""

from .definition import (SlotDefinition, )
from .rule_content import (DummySlotValRuleContent, SlotRuleContent)



"""
Module `chatette.cli.interactive_commands.set_modifier_command`.
Contains the strategy class that represents the interactive mode command
`set-modifier` which allows to change the value of the modifier of a unit
declaration.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.utils import str_to_bool
from chatette.parsing.parser_utils import CASE_GEN_SYM, ARG_SYM


class SetModifierCommand(CommandStrategy):
    usage_str = 'set-modifier <unit-type> "<unit-type>" ' + \
                '<modifer-name> "<value>"'

    def execute(self, facade):
        if len(self.command_tokens) < 5:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         self.usage_str)
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log("Unknown unit type: '" +
                                         str(self.command_tokens[1]) + "'.")
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])

        modifier_name = self.command_tokens[3]
        modifier_value = CommandStrategy.split_exact_unit_name(self.command_tokens[4])

        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(self.command_tokens[2])
            except SyntaxError:
                self.print_wrapper.error_log("Unit identifier couldn't be " + \
                                             "interpreted. Did you mean to " + \
                                             "escape some hashtags '#'?")
                return
            if variation_name is not None:
                self.print_wrapper.error_log("Cannot set a modifier for the " + \
                                             "variation of a unit.")
                return
            self._set_modifier(facade.parser, unit_type, unit_name,
                               modifier_name, modifier_value)
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(facade.parser,
                                                          unit_type,
                                                          unit_regex):
                self._set_modifier(facade.parser, unit_type, unit_name,
                                   modifier_name, modifier_value)
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")

    def _set_modifier(self, parser, unit_type, unit_name, modifier_name, value):
        unit = parser.get_definition(unit_name, unit_type)
        modifier_name = modifier_name.lower()
        if modifier_name in ("casegen", CASE_GEN_SYM):
            try:
                value = str_to_bool(value)
                unit.modifiers.casegen = value
            except ValueError:
                self.print_wrapper.write("Invalid value for case generation " +
                                         "modifier (True or False).")
                return
        elif modifier_name in ("arg", ARG_SYM):
            unit.set_arg(value)
        else:
            self.print_wrapper.write("Invalid modifier selected (can be "+
                                     "'casegen' or 'arg').")
            return
        self.print_wrapper.write("Modifier for " + unit_type.name + " '" +
                                 unit_name + "' successfully changed.")


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



"""
Module `chatette.cli.interactive_commands.declare_command`.
Contains the strategy class that represents the interactive mode command
`declare` which creates a new empty unit and add it to the list of units
of the parser.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.parsing.parser_utils import UnitType
from chatette.units.alias.definition import AliasDefinition
from chatette.units.slot.definition import SlotDefinition
from chatette.units.intent.definition import IntentDefinition


class DeclareCommand(CommandStrategy):
    usage_str = 'declare <unit-type> "<unit-name>"'

    def execute(self, facade):
        """
        Implements the command `rule` which generates a certain number of
        examples according to a provided rule.
        """
        if len(self.command_tokens) < 3:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         self.usage_str)
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log("Unknown unit type: '" +
                                         str(self.command_tokens[1]) + "'.")
            return

        try:
            [unit_name, variation_name] = \
                CommandStrategy.split_exact_unit_name(self.command_tokens[2])
        except SyntaxError:
            self.print_wrapper.error_log("Unit identifier couldn't be " + \
                                         "interpreted. Did you mean to " + \
                                         "escape some hashtags '#'?")
            return
        if variation_name is not None and variation_name != "":
            self.print_wrapper.error_log("Variation name detected, while " + \
                                         "units cannot be declared with a " + \
                                         "variation. Did you mean to escape " + \
                                         "some hashtags '#'?")
            return

        if unit_type == UnitType.alias:
            declaration = AliasDefinition(unit_name, None)
            relevant_dict = facade.parser.alias_definitions
        elif unit_type == UnitType.slot:
            declaration = SlotDefinition(unit_name, None)
            relevant_dict = facade.parser.slot_definitions
        else:  # intent
            declaration = IntentDefinition(unit_name, None)
            relevant_dict = facade.parser.intent_definitions

        if unit_name in relevant_dict:
            self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                     unit_name + "' is already defined.")
            return
        relevant_dict[unit_name] = declaration
        self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                 unit_name + "' was successfully declared.")


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



from chatette.units import Example, RuleContent, may_change_leading_case
from chatette.parsing.parser_utils import add_escapement_back_in_word


class WordRuleContent(RuleContent):
    """
    Represents a word inside a rule
    Accepted modifiers:
        - leading-space: bool
    """

    def __init__(self, name, leading_space=False, variation_name=None, arg_value=None,
                 casegen=False, randgen=None, percentage_gen=None, parser=None):

        if variation_name is not None:
            raise SyntaxError("Words cannot have variations, yet '" +
                              name + "' does (unescaped '#'?)")
        if arg_value is not None:
            raise SyntaxError("Words cannot have an argument, yet '" +
                              name + "' does (unescaped ':'?)")
        if casegen:
            raise SyntaxError("Words cannot generate different cases, yet '" +
                              name + "' does (unescaped '&'?)")
        if randgen is not None or percentage_gen is not None:
            raise SyntaxError("Words cannot have a random generation modifier, yet '" +
                              name + "' does (unescaped '?'?)")
        if parser is not None:
            raise RuntimeError("Internal error: tried to create a word " +
                               "with a pointer to the parser")

        super(WordRuleContent, self).__init__(name, leading_space=leading_space,
                                              variation_name=None, arg_value=None,
                                              casegen=False, randgen=None,
                                              percentage_gen=None, parser=None)
        self.word = name


    def can_have_casegen(self):
        return may_change_leading_case(self.word)

    def get_max_nb_generated_examples(self):
        return 1


    def generate_random(self, arg_value=None):
        if self.leading_space:
            return Example(' ' + self.word)

        return Example(self.word)

    def generate_all(self):
        if self.leading_space:
            return [Example(' ' + self.word)]

        return [Example(self.word)]


    def as_string(self):
        """
        Returns the representation of the rule
        as it would be written in a template file.
        """
        if self.leading_space:
            return ' ' + add_escapement_back_in_word(self.name)
        return add_escapement_back_in_word(self.name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module `chatette.utils`
Contains utility functions used everywhere in the project.
"""

from __future__ import print_function
import sys
from random import randint


# pylint: disable=invalid-name
def print_DBG(txt):
    """Prints debug information on stdout."""
    print("[DBG] " + txt)


def print_warn(txt):
    """Warns the user using stdout."""
    print("[WARN] " + txt)


def cast_to_unicode(anything):
    """
    If executed with Python 2.7, cast any string that is in `anything`
    to unicode.
    If executed with Python 3, returns `anything`.
    `anything` can be a string, an array, a dict,...
    """
    if sys.version_info[0] == 3:
        return anything

    if isinstance(anything, str):
        return unicode(anything, "utf-8")
    if isinstance(anything, dict):
        cast_dict = dict()
        for key in anything:
            cast_key = cast_to_unicode(key)
            cast_value = cast_to_unicode(anything[key])
            cast_dict[cast_key] = cast_value
        return cast_dict
    if isinstance(anything, list):
        cast_list = []
        for e in anything:
            cast_list.append(cast_to_unicode(e))
        return cast_list
    return anything


def choose(array):
    """
    Same as `random.choice(array)` but doesn't throw an error
    if `array` is empty
    """
    # ([anything]) -> anything or None
    array_len = len(array)
    if array_len <= 0:
        return None  # None and not [] (it is used later on in the code)
    return array[randint(0, array_len - 1)]


def rchop(string, ending):
    """Removes a substring at the end of a string."""
    if string.endswith(ending):
        return string[:-len(ending)]
    return string


def str_to_bool(text):
    """
    Transforms the strings 'True' and 'False' to their boolean counterparts.
    Raises a `ValueError` if `text` is neither of them.
    """
    text = text.lower()
    if text == "true":
        return True
    if text == "false":
        return False
    raise ValueError("Cannot convert '" + str(text) + "' into a boolean")


def remove_duplicates(dict_of_lists):
    """Removes duplicates from a dictionary containing lists."""
    return {key: list(set(value)) for (key, value) in dict_of_lists.items()}



if __name__ == "__main__":
    # pylint: disable=wrong-import-position
    import warnings

    warnings.warn("You are running the wrong file ('utils.py')." +
                  "The file that should be run is 'run.py'.")



import random
#from typing import List

from chatette.units import UnitDefinition#, Example
from .example import IntentExample


class IntentDefinition(UnitDefinition):
    """
    This class represents the definition of an intent,
    containing all the rules it can generate from.
    """

    def __init__(self, name, modifiers, rules=None):
        super(IntentDefinition, self).__init__(name, modifiers, rules=rules)

        self.type = "intent"
        self.nb_training_examples_asked = None  # All possibilities will be generated TODO
        self.nb_testing_examples_asked = None

    def set_nb_examples_asked(self, nb_training_examples_asked, nb_testing_examples_asked=None):
        self.nb_training_examples_asked = nb_training_examples_asked
        self.nb_testing_examples_asked = nb_testing_examples_asked


    def generate_random(self, variation_name=None, arg_value=None):
        example = \
            super(IntentDefinition, self).generate_random(variation_name, arg_value)
        tmp = IntentExample.from_example(self.name, example)
        return tmp
    
    def generate_all(self, variation_name=None, arg_value=None):
        examples = \
            super(IntentDefinition, self).generate_all(variation_name, arg_value)
        return [IntentExample.from_example(self.name, ex) for ex in examples]

    def generate(self, max_nb_examples, training_examples=None):# -> List[Example]:
        """
        Generates all the examples that were asked (i.e. as much examples
        as asked). The number of generated examples is tied to a maximum though TODO.
        When `training_examples` is `None`, this will generate the training examples
        (i.e. the number of training examples asked); otherwise, it will generate
        examples that are not in `training_examples` (if possible).
        """
        if training_examples is None and self.nb_training_examples_asked is None:
            return [
                    IntentExample(self.name, ex.text.strip(), ex.entities)
                    for (i, ex) in enumerate(self.generate_all())
                    if i < max_nb_examples
                ]

        nb_examples_asked = self.nb_training_examples_asked
        if training_examples is not None:
            if self.nb_testing_examples_asked is None:
                return []  # No examples must be generated
            nb_examples_asked = self.nb_testing_examples_asked

        if nb_examples_asked <= 0:
            return []

        nb_possible_ex = self.get_max_nb_generated_examples()
        if nb_examples_asked > nb_possible_ex:
            if training_examples is None:
                return [
                    IntentExample(self.name, ex.text.strip(), ex.entities)
                    for (i, ex) in enumerate(self.generate_all())
                    if i < max_nb_examples
                ]

            all_examples = [
                IntentExample(self.name, ex.text.strip(), ex.entities)
                for (i, ex) in enumerate(self.generate_all())
                if i < max_nb_examples
            ]
            return [
                ex
                for ex in all_examples if ex not in training_examples
            ]

        if nb_examples_asked > max_nb_examples:
            nb_examples_asked = max_nb_examples

        if nb_examples_asked < nb_possible_ex / 2:  # QUESTION: should this be /2?
            generated_examples = []
            for _ in range(nb_examples_asked):
                nb_iterations = 0
                while nb_iterations < 50:  # 50 is completely arbitrary
                    current_example = self.generate_random()
                    current_example.text = current_example.text.strip()  # Strip for safety
                    if (    current_example not in generated_examples
                        and (training_examples is None
                             or current_example not in training_examples)):
                        generated_examples.append(
                            IntentExample(self.name, current_example.text,
                                          current_example.entities))
                        break
                    nb_iterations += 1
            return generated_examples

        all_examples = [
            IntentExample(self.name, ex.text.strip(), ex.entities)
            for ex in self.generate_all()
        ]

        if training_examples is None:
            return random.sample(all_examples, nb_examples_asked)
        random.shuffle(all_examples)
        return [
            ex for ex in all_examples
            if ex not in training_examples
        ]


    def _get_template_decl(self, variation=None):
        result = '%' + \
                 super(IntentDefinition, self)._get_template_decl(variation)
        if self.nb_training_examples_asked is not None:
            result += "(train:" + str(self.nb_training_examples_asked)
            if self.nb_testing_examples_asked is not None:
                result += ", test:" + str(self.nb_testing_examples_asked) + ')'
        elif self.nb_testing_examples_asked is not None:
            result += "(test:" + str(self.nb_testing_examples_asked) +')'
        return result

    def short_desc_str(self):
        """
        Returns a str representing a short description of this unit description.
        """
        desc = super(IntentDefinition, self).short_desc_str() + '\n'
        if self.nb_training_examples_asked is None:
            desc += "# training examples: all\n"
        else:    
            desc += "# training examples: " + \
                    str(self.nb_training_examples_asked) + '\n'
        if self.nb_testing_examples_asked is None:
            desc += "# testing examples: all"
        else:
            desc += "# testing examples: " + str(self.nb_testing_examples_asked)
        return desc



"""
Module `chatette.parsing.new_parser`
Contains the (new) parser that reads and parses template files
and transforms the information they contain into dictionaries
of unit definitions.
"""

from six import string_types

from chatette.utils import print_DBG
import chatette.parsing.parser_utils as pu
from chatette.parsing.tokenizer import Tokenizer

from chatette.units.word.rule_content import WordRuleContent
from chatette.units.word.group_rule_content import GroupWordRuleContent
from chatette.units.choice import ChoiceRuleContent
from chatette.units.alias import AliasDefinition, AliasRuleContent
from chatette.units.slot import SlotDefinition, \
                                SlotRuleContent, DummySlotValRuleContent
from chatette.units.intent import IntentDefinition, IntentRuleContent


class Parser(object):
    """
    **This class is a refactor of the current parser**
    This class will parse the input file(s)
    and create an internal representation of its contents.
    """

    def __init__(self, master_filename):
        if not isinstance(master_filename, string_types):
            raise ValueError("Since v1.4.0, the parser takes as an argument " + \
                             "the path of the master file directly, rather " + \
                             "than the file itself as before.")
        self.tokenizer = Tokenizer(master_filename)

        self._expecting_rule = False
        self._currently_parsed_declaration = None  # 3-tuple
        self._expected_indentation = None  # str

        self.alias_definitions = dict()
        self.slot_definitions = dict()
        self.intent_definitions = dict()

        self.stats = {"#files": 1, "#declarations": 0, "#rules": 0,
                      "#intents": 0, "#aliases": 0, "#slots": 0}

    def open_new_master_file(self, master_filepath):
        self.tokenizer.redefine_master_file(master_filepath)
        self.stats["#files"] += 1


    def get_definition(self, definition_name, unit_type):
        """Returns the definition for unit with name `definition_name`."""
        if unit_type == pu.UnitType.alias:
            relevant_dict = self.alias_definitions
        elif unit_type == pu.UnitType.slot:
            relevant_dict = self.slot_definitions
        elif unit_type == pu.UnitType.intent:
            relevant_dict = self.intent_definitions
        else:
            raise ValueError("Tried to get a definition with wrong type "+
                             "(expected alias, slot or intent)")

        if definition_name not in relevant_dict:
            raise KeyError("Couldn't find a definition for "+unit_type.name+
                           " '"+definition_name+"' (did you mean to use "+
                           "the word group '["+definition_name+"]'?)")

        return relevant_dict[definition_name]


    def parse(self):
        """
        Parses the master file and subsequent files and
        transforms the information parsed into a dictionary of
        declaration names -> rules.
        """
        print_DBG("Parsing master file: "+self.tokenizer.get_file_information()[0])
        for token_line in self.tokenizer.next_tokenized_line():
            if not token_line[0].isspace():
                if token_line[0] == pu.INCLUDE_FILE_SYM:
                    self.tokenizer.open_file(token_line[1])
                    print_DBG("Parsing file: "+self.tokenizer.get_file_information()[0])
                    self.stats["#files"] += 1
                else:
                    self._parse_declaration_initiator(token_line)
                    self._expecting_rule = True
                    self.stats["#declarations"] += 1
                self._expected_indentation = None
            else:
                self._parse_rule(token_line)
                self._expecting_rule = False  # Not expecting but still allowed
                self.stats["#rules"] += 1
        self.tokenizer.close_files()
        print_DBG("Parsing finished!")

    def _parse_declaration_initiator(self, token_line):
        """Parses a line (as tokens) that contains a declaration initiator."""
        if self._expecting_rule:
            self.tokenizer.syntax_error("Expected a generation rule, got a "+
                                        "unit declaration instead.")

        unit_type = pu.get_unit_type_from_sym(token_line[0])

        declaration_interior = pu.get_declaration_interior(token_line)
        if declaration_interior is None:
            self.tokenizer.syntax_error("Couldn't find a valid unit declaration.")

        try:
            pu.check_declaration_validity(declaration_interior)
        except SyntaxError as e:
            self.tokenizer.syntax_error(str(e))

        unit_name = pu.find_name(declaration_interior)
        modifiers = pu.find_modifiers_decl(declaration_interior)
        nb_examples_asked = None
        if unit_type == pu.UnitType.intent:
            annotation_interior = pu.get_annotation_interior(token_line)
            if annotation_interior is not None and len(annotation_interior) > 0:
                nb_examples_asked = pu.find_nb_examples_asked(annotation_interior)

        self.create_unit(unit_type, unit_name, modifiers, nb_examples_asked)
        self._currently_parsed_declaration = (unit_type, unit_name, modifiers)

    def create_unit(self, unit_type, unit_name, modifiers,
                    nb_examples_asked=None):
        """
        Creates a unit of type `unit_type` with name `unit_name` and modifiers
        `modifiers` inside the relevant dictionary (`alias_definitions`,
        `slot_definitions` or `intent_definitions`).
        `modifiers` is a `UnitDeclarationModifiersRepr` and `nb_examples_asked`
        is a tuple (training, test).
        """
        new_unit = None
        relevant_dict = None
        if unit_type == pu.UnitType.alias:
            new_unit = AliasDefinition(unit_name, modifiers)
            # new_unit = AliasDefinition(unit_name, [], modifiers.argument_name,
            #                            modifiers.casegen)
            relevant_dict = self.alias_definitions
            self.stats["#aliases"] += 1
        elif unit_type == pu.UnitType.slot:
            new_unit = SlotDefinition(unit_name, modifiers)
            # new_unit = SlotDefinition(unit_name, [], modifiers.argument_name,
            #                           modifiers.casegen)
            relevant_dict = self.slot_definitions
            self.stats["#slots"] += 1
        elif unit_type == pu.UnitType.intent:
            new_unit = IntentDefinition(unit_name, modifiers)
            # new_unit = IntentDefinition(unit_name, [], modifiers.argument_name,
            #                             modifiers.casegen)
            relevant_dict = self.intent_definitions
            self.stats["#intents"] += 1

        if unit_type == pu.UnitType.intent and nb_examples_asked is not None:
            (train_nb, test_nb) = nb_examples_asked
            new_unit.set_nb_examples_asked(train_nb, test_nb)

        if unit_name not in relevant_dict:
            relevant_dict[unit_name] = new_unit
        elif modifiers.variation_name is None:
            pass  # Rules will be added to the already defined unit


    def _parse_rule(self, tokens):
        """
        Parses the list of string `tokens` that represents a rule in a
        template file and
        add the rule to the declaration currently being parsed.
        """
        if self._currently_parsed_declaration is None:
            self.tokenizer.syntax_error("Got a rule outside of "+
                                        "a unit declaration.")

        self._check_indentation(tokens[0])

        sub_rules = self.tokens_to_sub_rules(tokens[1:])

        relevant_dict = None
        if self._currently_parsed_declaration[0] == pu.UnitType.alias:
            relevant_dict = self.alias_definitions
        elif self._currently_parsed_declaration[0] == pu.UnitType.slot:
            relevant_dict = self.slot_definitions
        else:  # intent
            relevant_dict = self.intent_definitions

        name = self._currently_parsed_declaration[1]
        variation_name = self._currently_parsed_declaration[2].variation_name
        relevant_dict[name].add_rule(sub_rules, variation_name)

    def _check_indentation(self, indentation):
        """
        Checks that the str `indentation` is the same
        as the expected indentation from the last parsed rule.
        Raises a `SyntaxError` if it isn't.
        """
        if self._expected_indentation is None:
            self._expected_indentation = indentation
            return
        if indentation != self._expected_indentation:
            self.tokenizer.syntax_error("Inconsistent indentation.")

    def tokens_to_sub_rules(self, tokens):
        """Transforms a list of tokens into a list of sub-rules"""
        # Remove alternative slot rule if necessary
        alt_slot_value = None  # TODO DummySlotVal absolutely need to be managed differently
        if self._currently_parsed_declaration[0] == pu.UnitType.slot:
            answer = pu.find_alt_slot_and_index(tokens)
            if answer is not None:
                (end_index, alt_slot_value) = answer
                tokens = tokens[:end_index]

        sub_rules = []
        leading_space = False
        for sub_rule_tokens in pu.next_sub_rule_tokens(tokens):
            if len(sub_rule_tokens) == 1 and sub_rule_tokens[0] == ' ':
                leading_space = True
                continue

            sub_rule = self._make_sub_rule_from_tokens(sub_rule_tokens,
                                                       leading_space)
            if alt_slot_value is not None:
                if alt_slot_value != pu.ALT_SLOT_VALUE_FIRST_SYM:
                    sub_rules = [DummySlotValRuleContent(alt_slot_value, sub_rule)]
                else:
                    sub_rules = [DummySlotValRuleContent(sub_rule.name, sub_rule)]
                alt_slot_value = None
            sub_rules.append(sub_rule)

            leading_space = False
        return sub_rules

    def _make_sub_rule_from_tokens(self, sub_rule_tokens, leading_space):
        if pu.is_sub_rule_word(sub_rule_tokens):
            word_text = pu.remove_escapement(sub_rule_tokens[0])
            word = WordRuleContent(word_text, leading_space)
            return word
        if pu.is_sub_rule_word_group(sub_rule_tokens):
            group_interior_tokens = sub_rule_tokens[1:-1]
            self._check_sub_rule_validity(group_interior_tokens,
                                          pu.SubRuleType.word_group)
            words = pu.find_words(group_interior_tokens)
            words_str = pu.remove_escapement(''.join(words))
            modifiers = pu.find_modifiers_word_group(group_interior_tokens)
            word_group = \
                GroupWordRuleContent(words_str, leading_space,
                                     casegen=modifiers.casegen,
                                     randgen=modifiers.randgen_name,
                                     percentage_gen=modifiers.percentage_randgen)
            return word_group
        if pu.is_sub_rule_choice(sub_rule_tokens):
            choice_interior_tokens = sub_rule_tokens[1:-1]
            self._check_sub_rule_validity(choice_interior_tokens,
                                          pu.SubRuleType.choice)
            modifiers = pu.find_modifiers_choice(choice_interior_tokens)
            choice = ChoiceRuleContent(''.join(choice_interior_tokens),
                                       leading_space,
                                       casegen=modifiers.casegen,
                                       randgen=modifiers.randgen)#,
                                       #percentage_gen=modifiers.percentage_randgen)
            for choice_tokens in pu.next_choice_tokens(choice_interior_tokens):
                current_leading_space = False
                current_choice_sub_rules = []
                for choice_sub_rule_tokens in pu.next_sub_rule_tokens(choice_tokens):
                    if choice_sub_rule_tokens == ' ':
                        current_leading_space = True
                        continue
                    choice_sub_rule = \
                        self._make_sub_rule_from_tokens(choice_sub_rule_tokens,
                                                        current_leading_space)
                    current_choice_sub_rules.append(choice_sub_rule)
                    current_leading_space = False
                choice.add_choice(current_choice_sub_rules)
            return choice

        if (   len(sub_rule_tokens) == 0
            or not pu.is_unit_type_sym(sub_rule_tokens[0])):
            # Not any type of sub-rule detected
            self.tokenizer.syntax_error("Invalid type of sub-rule.",
                                        word_to_find=sub_rule_tokens[0])

        interior_tokens = sub_rule_tokens[2:-1]
        name = pu.find_name(interior_tokens)
        if pu.is_sub_rule_alias_ref(sub_rule_tokens):
            self._check_sub_rule_validity(interior_tokens,
                                          pu.SubRuleType.alias)
            modifiers = pu.find_modifiers_reference(interior_tokens)
            alias = AliasRuleContent(name, leading_space,
                                     modifiers.variation_name,
                                     modifiers.argument_value,
                                     modifiers.casegen,
                                     modifiers.randgen_name,
                                     modifiers.percentage_randgen, self)
            return alias
        if pu.is_sub_rule_slot_ref(sub_rule_tokens):
            self._check_sub_rule_validity(interior_tokens,
                                          pu.SubRuleType.slot)
            modifiers = pu.find_modifiers_reference(interior_tokens)
            slot = SlotRuleContent(name, leading_space,
                                   modifiers.variation_name,
                                   modifiers.argument_value,
                                   modifiers.casegen,
                                   modifiers.randgen_name,
                                   modifiers.percentage_randgen, self)
            return slot
        if pu.is_sub_rule_intent_ref(sub_rule_tokens):
            self._check_sub_rule_validity(interior_tokens,
                                          pu.SubRuleType.intent)
            modifiers = pu.find_modifiers_reference(interior_tokens)
            intent = IntentRuleContent(name, leading_space,
                                       modifiers.variation_name,
                                       modifiers.argument_value,
                                       modifiers.casegen,
                                       modifiers.randgen_name,
                                       modifiers.percentage_randgen, self)
            return intent

    def _check_sub_rule_validity(self, interior_tokens, sub_rule_type):
        """
        Checks that the sub-rule represented by `interior_tokens` is
        syntactically valid.
        """
        try:
            if sub_rule_type == pu.SubRuleType.word_group:
                pu.check_word_group_validity(interior_tokens)
            elif sub_rule_type == pu.SubRuleType.choice:
                pu.check_choice_validity(interior_tokens)
            elif sub_rule_type in (pu.SubRuleType.alias,
                                   pu.SubRuleType.slot,
                                   pu.SubRuleType.intent):
                pu.check_reference_validity(interior_tokens)
        except SyntaxError as e:
            self.tokenizer.syntax_error(str(e))


    def rename_unit(self, unit_type, old_name, new_name):
        """
        Renames the unit declaration of type `unit_type` from
        `old_name` to `new_name` (possibly replacing the unit with that name).
        Raises a `KeyError` if `old_name` is not a declared unit.
        @post: this can lead to inconsistent rules.
        """
        if unit_type == pu.UnitType.alias:
            relevant_dict = self.alias_definitions
        elif unit_type == pu.UnitType.slot:
            relevant_dict = self.slot_definitions
        elif unit_type == pu.UnitType:
            relevant_dict = self.intent_definitions
        else:
            raise ValueError("Tried to rename a definition with wrong type "+
                             "(expected alias, slot or intent)")

        if old_name in relevant_dict:
            if new_name in relevant_dict:
                raise ValueError("Tried to rename a definition to a name that " +
                                 "was already in use ('" + new_name + "').")
            relevant_dict[new_name] = relevant_dict[old_name]
            del relevant_dict[old_name]
            relevant_dict[new_name].name = new_name
        else:
            raise KeyError("No unit named '"+old_name+"' was found")

    def delete(self, unit_type, unit_name, variation_name=None):
        """Deletes a unit definition."""
        if unit_type == pu.UnitType.alias:
            relevant_dict = self.alias_definitions
            stat_key = "#aliases"
        elif unit_type == pu.UnitType.slot:
            relevant_dict = self.slot_definitions
            stat_key = "#slots"
        elif unit_type == pu.UnitType.intent:
            relevant_dict = self.intent_definitions
            stat_key = "#intents"
        else:
            raise ValueError("Tried to delete a definition with wrong type "+
                             "(expected alias, slot or intent)")

        if unit_name not in relevant_dict:
            raise KeyError("Couldn't find a definition for " + unit_type.name +
                           " '" + unit_name + "'.")

        nb_rules = relevant_dict[unit_name].get_nb_rules(variation_name)
        if variation_name is None:
            del relevant_dict[unit_name]
            self.stats[stat_key] -= 1
            self.stats["#declarations"] -= 1
            self.stats["#rules"] -= nb_rules
        else:
            relevant_dict[unit_name].delete_variation(variation_name)
            self.stats["#rules"] -= nb_rules


    def add_definition(self, unit_type, unit_name, definition):
        """Adds an already built definition to the list of declared units."""
        if unit_type == pu.UnitType.alias:
            relevant_dict = self.alias_definitions
            stat_key = "#aliases"
        elif unit_type == pu.UnitType.slot:
            relevant_dict = self.slot_definitions
            stat_key = "#slots"
        elif unit_type == pu.UnitType.intent:
            relevant_dict = self.intent_definitions
            stat_key = "#intents"
        else:
            raise ValueError("Tried to delete a definition with wrong type "+
                             "(expected alias, slot or intent)")

        if unit_name in relevant_dict:
            raise ValueError(unit_type.name.capitalize()+" '"+unit_name+"' " +
                             "is already defined. Tried to add a definition " +
                             "for it again.")

        relevant_dict[unit_name] = definition
        self.stats[stat_key] += 1
        self.stats["#declarations"] += 1
        self.stats["#rules"] += definition.get_nb_rules()


    def print_DBG(self):
        print("Aliases:")
        for alias_name in self.alias_definitions:
            self.alias_definitions[alias_name].print_DBG()
        print("Slots:")
        for slot_name in self.slot_definitions:
            self.slot_definitions[slot_name].print_DBG()
        print("Intents:")
        for intent_name in self.intent_definitions:
            self.intent_definitions[intent_name].print_DBG()
        print()



"""
Module `chatette.cli.interactive_commands.delete_command`.
Contains the strategy class that represents the interactive mode command
`delete` which deletes a unit declaration from the parser.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class DeleteCommand(CommandStrategy):
    usage_str = 'delete <unit-type> "<unit-name>"'
    def __init__(self, command_str, quiet=False):
        super(DeleteCommand, self).__init__(command_str, quiet)
        self._units_to_delete = []

    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        self._units_to_delete.append((unit_type, unit_name, variation_name))
        self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                unit_name + "' was successfully deleted.")

    def finish_execution(self, facade):
        for (unit_type, unit_name, variation_name) in self._units_to_delete:
            try:
                facade.parser.delete(unit_type, unit_name, variation_name)
            except KeyError:
                self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                        unit_name + "' was not defined.")
        self._units_to_delete = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from warnings import warn

from chatette.utils import print_warn


# ============ Deprecation of semi-colon syntax for comments ===============
# Comments starting with semi-colons ';' are now deprecated to have a closer
# syntax to *Chatito* v2.1.x
_SEMICOLON_COMMENTS_DEPRECATION_WARNED = False


def warn_semicolon_comments():
    """
    Warns the user on stdout that one of their files contains semicolons
    comments (which are a deprecated way of making comments).
    Rather use '//' comments instead of ';' comments.
    """
    global _SEMICOLON_COMMENTS_DEPRECATION_WARNED
    if not _SEMICOLON_COMMENTS_DEPRECATION_WARNED:
        # pylint: disable=bad-continuation
        print_warn("Comments starting with a semi-colon ';' are "+
                   "now deprecated. Rather use the new double slash '//'"+
                   " syntax. This syntax allows to have a syntax closer to "+
                   "Chatito v2.1.x.")
        warn("Comments starting with a semi-colon ';' are now deprecated. " +
             "Rather use the new double slash '//' syntax. This " +
             "syntax allows to have a syntax closer to Chatito v2.1.x.",
             DeprecationWarning)
        _SEMICOLON_COMMENTS_DEPRECATION_WARNED = True



"""
Module `chatette.cli.interactive_commands.show_command`.
Contains the strategy class that represents the interactive mode command
`show` which shows information about a unit definition and lists a bunch of
its rules (all if possible).
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class ShowCommand(CommandStrategy):
    usage_str = 'show <unit-type> "<unit-name>"'
    max_nb_rules_to_display = 12

    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        try:
            unit = facade.parser.get_definition(unit_name, unit_type)
            self.print_wrapper.write(unit.short_desc_str())

            if variation_name is None:
                rules = unit.rules
                self.print_wrapper.write("Rules:")
            else:
                if variation_name not in unit.variations:
                    self.print_wrapper.error_log("Variation '" + variation_name + \
                                                 "' is not defined in " + \
                                                 unit.type + " " + unit_name + '.')
                    return
                rules = unit.variations[variation_name]
                self.print_wrapper.write("Rules for variation '" + \
                                         variation_name + "':")
            
            for (i, rule) in enumerate(rules):
                if i >= self.max_nb_rules_to_display:
                    break
                rule_str = ''.join([sub_rule.as_string() for sub_rule in rule])
                self.print_wrapper.write('\t'+rule_str)
            self.print_wrapper.write("")
        except KeyError:
            self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                     unit_name + "' is not defined.")



"""
Module `chatette.cli.interactive_commands.generate_command`.
Contains the strategy class that represents the interactive mode command
`generate` which generates units and writes them out formatted using a
given adapter.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.adapters.factory import create_adapter


class GenerateCommand(CommandStrategy):
    usage_str = 'generate <adapter> <unit-type> "<unit-name>"'
    def __init__(self, command_str, quiet=False):
        super(GenerateCommand, self).__init__(command_str, quiet)
        self.nb_examples = None

    def execute(self, facade):
        """
        Implements the command `generate` which generates all possible examples
        of a certain unit, formatted according to a certain adapter.
        """
        if len(self.command_tokens) == 1:
            facade.run_generation()
            return
        if len(self.command_tokens) == 2:
            try:
                facade.run_generation(self.command_tokens[1])
            except ValueError:
                self.print_wrapper.write("Unknown adapter: '" +
                                         self.command_tokens[1] + "'")
            return
        if len(self.command_tokens) < 4:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         self.usage_str)
            return

        adapter_str = self.command_tokens[1]
        try:
            adapter = create_adapter(adapter_str)
        except ValueError:
            self.print_wrapper.error_log("Unknown adapter '" + adapter_str + \
                                         "'.")

        if len(self.command_tokens) == 5:
            try:
                self.nb_examples = int(self.command_tokens[-1])
            except ValueError:
                self.print_wrapper.error_log("The number of examples to be " +
                                             "generated is invalid: it must " +
                                             "be an integer (no other " +
                                             "characters allowed).")
                return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[2])
        if unit_type is None:
            self.print_wrapper.error_log("Unknown unit type: '" +
                                         str(self.command_tokens[2]) + "'.")
            return

        unit_regex = self.get_regex_name(self.command_tokens[3])
        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(self.command_tokens[3])
            except SyntaxError:
                self.print_wrapper.error_log("Unit identifier couldn't be " + \
                                             "interpreted. Did you mean to " + \
                                             "escape some hashtags '#'?")
                return
            self._generate_unit(facade, adapter, unit_type, unit_name,
                                variation_name)
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(facade.parser,
                                                          unit_type,
                                                          unit_regex):
                self._generate_unit(facade, adapter, unit_type, unit_name)
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")
        self.finish_execution(facade)

    def _generate_unit(self, facade, adapter, unit_type, unit_name,
                       variation_name=None):
        definition = facade.parser.get_definition(unit_name, unit_type)
        examples = definition.generate_nb_examples(self.nb_examples,
                                                   variation_name)

        self.print_wrapper.write("Generated examples for " + unit_type.name +
                                 " '" + unit_name + "':'")
        for ex in examples:
            # HACK: add a name of intent as ex is not especially a IntentExample
            if not hasattr(ex, 'name'):
                ex.name = "INTERACTIVE"
            self.print_wrapper.write(adapter.prepare_example(ex))
        self.print_wrapper.write("")

    def finish_execution(self, facade):
        self.nb_examples = None


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import sys

from chatette import __version__
from chatette.facade import Facade
from chatette.cli.interpreter import CommandLineInterpreter


def main():
    # pylint: disable=bad-continuation
    argument_parser = argparse.ArgumentParser(
        description="Chatette v"+__version__+" -- " +
                    "Generates NLU datasets from template files",
        epilog="SimGus -- 2018 -- Released under MIT license",
        prog="Chatette",
        add_help=True)

    argument_parser.add_argument("input", type=str,
                                 help="Path to master template file")

    argument_parser.add_argument("-v", "--version", action="version",
                                 version="%(prog)s v"+__version__,
                                 help="Print the version number of the module")

    argument_parser.add_argument("-o", "--out", dest="output", required=False,
                                 type=str, default=None,
                                 help="Output directory path")
    argument_parser.add_argument("-l", "--local", dest="local", required=False,
                                 action="store_true", default=False,
                                 help="Change the base directory for output " +
                                      "files from the current working directory " +
                                      "to the directory containing the template " +
                                      "file")

    argument_parser.add_argument("-a", "--adapter", dest="adapter", required=False,
                                 type=str, default="rasa",
                                 help="Write adapter. Possible values: " +
                                      "['rasa', 'jsonl']")
    argument_parser.add_argument("--base-file", dest="base_filepath", required=False,
                                type=str, default=None,
                                help="Path to base file to extend with examples " +
                                     "and synonyms. Only with Rasa adapter.")

    argument_parser.add_argument("-s", "--seed", dest="seed", required=False,
                                 type=str, default=None,
                                 help="Seed for the random generator " +
                                      "(any string without spaces will work)")
                                      
    argument_parser.add_argument("-i", "--interactive", dest="interactive_mode",
                                 required=False, action="store_true",
                                 default=False,
                                 help="Runs Chatette in interactive mode")
    argument_parser.add_argument("-I", "--interactive-commands-file",
                                 dest="interactive_commands_file",
                                 required=False, default=None, type=str,
                                 help="Path to a file containing interactive " +
                                      "mode commands that will be directly run")
    
    argument_parser.add_argument("-f", "--force", dest="force", required=False,
                                 action="store_true", default=False,
                                 help="Don't ask for confirmation before " +
                                      "overwriting files and folders")

    if len(sys.argv[1:]) == 0:
        argument_parser.print_help()
        argument_parser.exit()

    args = argument_parser.parse_args()

    facade = Facade.get_or_create_from_args(args)
    if not args.interactive_mode and args.interactive_commands_file is None:
        facade.run()
    else:
        cli = CommandLineInterpreter(facade, args.interactive_commands_file)
        cli.wait_for_input()


if __name__ == "__main__":
    main()



import json

from chatette.units import ENTITY_MARKER
from chatette.utils import cast_to_unicode
from ._base import Adapter


class RasaAdapter(Adapter):
    def __init__(self, base_filepath=None, batch_size=10000):# -> None:
        super(RasaAdapter, self).__init__(base_filepath, batch_size)
        self._base_file_contents = None

    def _get_file_extension(self):
        return "json"


    def prepare_example(self, example):
        def entity_to_rasa(entity):
            first_index = self.__find_entity(example.text, entity["text"])
            entity["text"] = entity["text"].rstrip()
            # NOTE: This always finds something
            # Remove the entity marker of this entity
            # (works unless entities are not recorded in order)
            example.text = example.text[:first_index] + \
                            example.text[first_index+len(ENTITY_MARKER):]
            return {
                "value": entity["value"],
                "entity": entity["slot-name"],
                "start": first_index,
                "end": first_index + len(entity["text"]),
            }

        return {
            "intent": example.name,
            "entities": [entity_to_rasa(e) for e in example.entities],
            # HACK: Call `entity_to_rasa` BEFORE storing "text" here
            #       (this function removes the entity markers)
            "text": example.text,
        }


    def _write_batch(self, output_file_handle, batch):
        rasa_entities = [self.prepare_example(ex) for ex in batch.examples]

        json_data = self._get_base_to_extend()
        json_data["rasa_nlu_data"]["common_examples"] = rasa_entities
        json_data["rasa_nlu_data"]["entity_synonyms"] = \
            self.__synonym_format(batch.synonyms)
        json_data = cast_to_unicode(json_data)

        output_file_handle.write(json.dumps(json_data, ensure_ascii=False,
                                            indent=2, sort_keys=True))

    @classmethod
    def __synonym_format(cls, synonyms):
        # {str: [str]} -> [{"value": str, "synonyms": [str]}]
        return [
            {
                "value": slot_name,
                "synonyms": synonyms[slot_name]
            }
            for slot_name in synonyms
            if len(synonyms[slot_name]) > 1
        ]

    @staticmethod
    def __find_entity(text, entity_str):
        """
        Finds the entity `entity_str` in `text`
        ignoring the case of the first non-space.
        """
        index = text.find((ENTITY_MARKER+entity_str).rstrip())
        if index == -1:
            return text.lower().find(entity_str.lower())
        return index


    def _get_base_to_extend(self):
        if self._base_file_contents is None:
            if self._base_filepath is None:
                return self._get_empty_base()
            with open(self._base_filepath, 'r') as base_file:
                self._base_file_contents = json.load(base_file)
            self.check_base_file_contents()
        return self._base_file_contents

    def _get_empty_base(self):
        return {
                    "rasa_nlu_data": {
                        "common_examples": None,
                        "regex_features": [],
                        "lookup_tables": [],
                        "entity_synonyms": None,
                    }
                }

    def check_base_file_contents(self): 
        """
        Checks that `self._base_file_contents` contains well formatted JSON.
        Throws a `SyntaxError` if the data is incorrect.
        """
        if self._base_file_contents is None:
            return
        if not isinstance(self._base_file_contents, dict):
            self._base_file_contents = None
            raise SyntaxError(
                "Couldn't load data from base file '" + \
                self._base_filepath + "'")
        else:
            if "rasa_nlu_data" not in self._base_file_contents:
                self._base_file_contents = None
                raise SyntaxError(
                    "Expected 'rasa_nlu_data' as a root of base file '" + \
                    self._base_filepath + "'")
            elif "common_examples" not in self._base_file_contents["rasa_nlu_data"] or \
                 "entity_synonyms" not in self._base_file_contents["rasa_nlu_data"]:
                    self._base_file_contents = None
                    raise SyntaxError(
                        "Expected at least 'entity_synonyms' and " + \
                        "'common_examples' inside 'rasa_nlu_data' " + \
                        "in base file '" + self._base_filepath + "'")



"""
Module `chatette.units.intent`
Contains the definition and the content rules charcterizing intents.
"""

from .definition import (IntentDefinition, )
from .example import (IntentExample, )
from .rule_content import (IntentRuleContent, )



from random import randint

from chatette.units import Example, RuleContent, may_change_leading_case, \
                           may_get_leading_space, randomly_change_case, \
                           with_leading_lower, with_leading_upper
from chatette.parsing.parser_utils import add_escapement_back_in_group


class GroupWordRuleContent(RuleContent):
    """
    Represents a word group token inside a rule.
    `name` is actually the words as a string.
    Accepted modifiers:
        - leading-space: bool
        - casegen: bool
        - randgen: str
        - percentgen: bool
    """

    def __init__(self, name, leading_space=False, variation_name=None, arg_value=None,
                 casegen=False, randgen=None, percentage_gen=50, parser=None):
        if variation_name is not None:
            raise SyntaxError("Word groups cannot have variations, yet '" +
                              name + "' does (unescaped '#'?)")
        if arg_value is not None:
            raise SyntaxError("Word groups cannot have an argument, yet '" +
                              name + "' does (unescaped '$'?)")
        if parser is not None:
            raise RuntimeError("Internal error: tried to create a word " +
                               "group with a pointer to the parser")

        if not may_change_leading_case(name):
            casegen = False

        super(GroupWordRuleContent, self).__init__(name,
                                                   leading_space=leading_space,
                                                   variation_name=None,
                                                   arg_value=None,
                                                   casegen=casegen,
                                                   randgen=randgen,
                                                   percentage_gen=percentage_gen,
                                                   parser=None)
        self.words = name

    def can_have_casegen(self):
        return may_change_leading_case(self.words)

    def generate_random(self, generated_randgens=None):
        if generated_randgens is None:
            generated_randgens = dict()

        if self.randgen is not None and self.randgen in generated_randgens:
            if generated_randgens[self.randgen]:
                pass  # Must be generated
            else:
                return Example()  # Cannot be generated
        elif self.randgen is not None:
            if randint(0, 99) >= self.percentgen:
                # Don't generated this randgen
                generated_randgens[self.randgen] = False
                return Example()
            else:
                # Generate this randgen
                generated_randgens[self.randgen] = True

        # Generate the string according to the parameters of the object
        generated_str = self.words
        if self.casegen:
            generated_str = randomly_change_case(generated_str)

        if self.leading_space and may_get_leading_space(generated_str):
            generated_str = ' ' + generated_str

        return Example(generated_str)

    def generate_all(self):
        generated_examples = []
        if self.randgen is not None:
            generated_examples.append("")

        if self.casegen:
            generated_examples.append(with_leading_lower(self.words))
            generated_examples.append(with_leading_upper(self.words))
        else:
            generated_examples.append(self.words)

        if self.leading_space:
            for (i, ex) in enumerate(generated_examples):
                if may_get_leading_space(ex):
                    generated_examples[i] = ' ' + ex

        result = [Example(ex) for ex in generated_examples]
        return result

    def get_max_nb_generated_examples(self):
        nb_possible_ex = 1
        if self.casegen:
            nb_possible_ex *= 2
        if self.randgen is not None:
            nb_possible_ex += 1
        return nb_possible_ex

    def as_string(self):
        """
        Returns the representation of the rule
        as it would be written in a template file.
        """
        result = add_escapement_back_in_group(self.name)
        if self.casegen:
            result = '&'+result
        if self.variation_name is not None:
            result += '#'+self.variation_name
        if self.randgen is not None:
            result += '?'+str(self.randgen)
            if self.percentgen != 50:
                result += '/'+str(self.percentgen)
        if self.arg_value is not None:
            result += '$'+self.arg_value
        result = '[' + result + ']'
        if self.leading_space:
            result = ' '+result
        return result



"""
Module `chatette.cli.interactive_commands.hide_command`.
Contains the strategy class that represents the interactive mode command
`hide` which hides unit definitions (storing them somewhere to be able to
unhide them later).
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class HideCommand(CommandStrategy):
    usage_str = 'hide <unit-type> "<unit-name>"'
    stored_units = {"alias": dict(), "slot": dict(), "intent": dict()}
    stored_variations = {"alias": dict(), "slot": dict(), "intent": dict()}
    def __init__(self, command_str, quiet=False):
        super(HideCommand, self).__init__(command_str, quiet)
        self._units_to_delete = []
        self._var_to_delete = []
    
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        try:
            unit = facade.parser.get_definition(unit_name, unit_type)
            if variation_name is None:
                self.stored_units[unit_type.name][unit_name] = unit
                self._units_to_delete.append((unit_type, unit_name))
                self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                        unit_name + "' was successfully hidden.")
            else:
                if variation_name not in unit.variations:
                    self.print_wrapper.error_log("Couldn't find variation '" + \
                                                 variation_name + "' in " + \
                                                 unit_type.name + " '" + \
                                                 unit_name + "'.")
                    return
                self._var_to_delete.append((unit_type, unit_name, variation_name))
                rules = unit.variations[variation_name]
                if unit_name not in self.stored_variations[unit_type.name]:
                    self.stored_variations[unit_type.name][unit_name] = \
                        {variation_name: rules}
                else:
                    self.stored_variations[unit_type.name][unit_name][variation_name] = \
                        rules
                self.print_wrapper.write("Variation '" + variation_name + "' of " + \
                                         unit_type.name + " '" + unit_name + \
                                         "' was successfully hidden.")
        except KeyError:
            self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                     unit_name + "' was not defined.")
    
    def finish_execution(self, facade):
        for (unit_type, unit_name) in self._units_to_delete:
            facade.parser.delete(unit_type, unit_name)
        self._units_to_delete = []

        for (unit_type, unit_name, variation_name) in self._var_to_delete:
            facade.parser.delete(unit_type, unit_name, variation_name)
        self._var_to_delete = []



"""
Module `chatette.parsing.tokenizer`
Contains the tokenizer used by the parser.
"""

import os

import chatette.parsing.parser_utils as pu
from chatette.parsing.line_count_file_wrapper import LineCountFileWrapper


class Tokenizer(object):
    def __init__(self, master_filename):
        self.master_file_paths = [master_filename]  # TODO use this
        self.master_file_dir = os.path.dirname(master_filename)
        
        self.current_file = LineCountFileWrapper(master_filename)
        self._opened_files = []

        self._last_read_line = None

    def redefine_master_file(self, master_filename):
        """
        Change the master file and open it.
        If other files are still open, raise an exception.
        """
        if len(self._opened_files) > 0:
            raise ValueError("Tried to change master file during parsing of " +
                             "other files")
        self.master_file_paths.append(master_filename)
        self.master_file_dir = os.path.dirname(master_filename)
        self.current_file = LineCountFileWrapper(master_filename)

    def open_file(self, filename):
        """
        Stores the current file for future use and opens `filename`.
        `filename` is given relatively to the master file.
        """
        filepath = os.path.join(os.path.dirname(self.current_file.name), filename)  # TODO this fix might be improved thanks to master_file_paths
        opened_filepaths = [f.name for f in self._opened_files]
        if filepath in opened_filepaths:
            raise ValueError("Tried to parse file '" + filepath + "' several " + \
                             "times (last time when parsing '" + \
                             self.current_file.name + \
                             "'). There seems to be circular includes in " + \
                             "template files.")
        if self.current_file is not None:
            self._opened_files.append(self.current_file)
        self.current_file = LineCountFileWrapper(filepath)

    def close_files(self):
        for f in self._opened_files:
            if not f.closed:
                f.close()
        if self.current_file is not None and not self.current_file.closed:
            self.current_file.close()

    def close_current_file(self):
        if self.current_file is not None:
            self.current_file.close()
        if len(self._opened_files) > 0:
            self.current_file = self._opened_files.pop()
        else:
            self.current_file = None

    def get_file_information(self):
        return (self.current_file.name, self.current_file.line_nb)


    def fail(self, exception):
        """Closes all files before raising an exception."""
        self.close_files()
        raise exception
    def syntax_error(self, message, line=None, line_index=0, word_to_find=None):
        """Makes an exception to be raised after closing all files."""
        if line is None:
            line = self._last_read_line
        if word_to_find is not None:
            line_index = line.find(word_to_find)
        exception = SyntaxError(message, (self.current_file.name,
                                          self.current_file.line_nb,
                                          line_index, line))
        self.fail(exception)


    def read_line(self):
        """
        Reads a line of `self.current_file` and returns it without the trailing
        new line (`\n`).
        If the file was entirely read, closes it and continues to read the
        file that was previously being read (returning its next line).
        Returns `None` if there is no file left to read.
        """
        line = self.current_file.readline()
        while line == '':  # EOF
            self.close_current_file()
            if self.current_file is None:  # No more files to read
                return None
            line = self.current_file.readline()
        line = line.rstrip()
        self._last_read_line = line
        return line

    def next_tokenized_line(self):
        """
        Yields the next relevant line of the current file as a list of tokens.
        An irrelevant line is an empty or comment line.
        """
        while True:
            line_str = pu.strip_comments(self.read_line())
            if line_str is None:
                break
            if line_str == "":
                continue
            yield self.tokenize(line_str)

    def tokenize(self, text):
        """
        Returns a tokenized version of the string `text`,
        i.e. a list of strings that make up words or special characters.
        The string `~[alias?] word. [&group]` would be tokenized into
        `["~", "[", "alias", "?", "]", " ", "word.", "[", "&", "group", "]"]`.
        @pre: `text` is not `None` or ''.
        """
        tokens = []
        current_token = ""

        def store_current_token():
            if current_token != "":
                tokens.append(current_token)
            return ""

        indentation = Tokenizer._get_indentation(text)
        if indentation is not None:
            tokens.append(indentation)
            text = text.lstrip()

        if indentation is None and text[0] == pu.INCLUDE_FILE_SYM:
            return [pu.INCLUDE_FILE_SYM, text[1:]]

        nb_closing_brackets_expected = 0  # For unit declaration initiators and references
        expecting_percent_gen = False  # True in a sub-rule after a `?`
        after_unit_declaration = False
        inside_annotation = False  # For parentheses after declaration initiator
        inside_choice = False
        next_char_escaped = False

        i = 0
        for (i, c) in enumerate(text):
            # Escapement
            if next_char_escaped:
                current_token += c
                next_char_escaped = False
            elif c == pu.ESCAPE_SYM:
                current_token += c  # Remove escapement later on
                next_char_escaped = True
            # Unit special characters
            elif c == pu.ALIAS_SYM:
                current_token = store_current_token()
                tokens.append(c)
            elif c == pu.SLOT_SYM:
                current_token = store_current_token()
                tokens.append(c)
            elif c == pu.INTENT_SYM:
                current_token = store_current_token()
                tokens.append(c)
            # Unit brackets
            elif c == pu.UNIT_OPEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
                nb_closing_brackets_expected += 1
            elif c == pu.UNIT_CLOSE_SYM:
                if nb_closing_brackets_expected < 1:
                    self.syntax_error("Inconsistent use of unit brackets "+
                                      "(too many closing unit symbols '"+
                                      pu.UNIT_CLOSE_SYM+"').", text, i)
                current_token = store_current_token()
                tokens.append(c)
                nb_closing_brackets_expected -= 1
                if indentation is None and nb_closing_brackets_expected == 0:
                    after_unit_declaration = True
            # Choice
            elif c == pu.CHOICE_OPEN_SYM:
                if inside_choice:
                    self.syntax_error("Nested choices are not supported. "+
                                      "Did you mean to escape it ('"+
                                      pu.ESCAPE_SYM+pu.CHOICE_OPEN_SYM+
                                      "' instead of '"+pu.CHOICE_OPEN_SYM+"'?",
                                      text, i)
                current_token = store_current_token()
                tokens.append(c)
                inside_choice = True
            elif c == pu.CHOICE_CLOSE_SYM:
                if not inside_choice:
                    self.syntax_error("Cannot close a choice before "+
                                      "opening it. Did you mean to escape "+
                                      "it ('"+pu.ESCAPE_SYM+
                                      pu.CHOICE_CLOSE_SYM+"' instead of '"+
                                      pu.CHOICE_CLOSE_SYM+"'?",
                                      text, i)
                current_token = store_current_token()
                tokens.append(c)
                inside_choice = False
            elif inside_choice and c == pu.CHOICE_SEP:
                current_token = store_current_token()
                tokens.append(c)
            # Inside unit
            elif nb_closing_brackets_expected > 0 and c == pu.VARIATION_SYM:
                current_token = store_current_token()
                tokens.append(c)
            elif nb_closing_brackets_expected > 0 and c == pu.RAND_GEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
                expecting_percent_gen = True
            elif nb_closing_brackets_expected > 0 and expecting_percent_gen and c == pu.PERCENT_GEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
                expecting_percent_gen = False
            elif nb_closing_brackets_expected > 0 and c == pu.CASE_GEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
            elif nb_closing_brackets_expected > 0 and c == pu.ARG_SYM:
                current_token = store_current_token()
                tokens.append(c)
            # Inside choice
            elif inside_choice and c == pu.RAND_GEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
            elif inside_choice and c == pu.CASE_GEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
            # Slot alternative value
            elif nb_closing_brackets_expected == 0 and c == pu.ALT_SLOT_VALUE_NAME_SYM:
                current_token = store_current_token()
                tokens.append(c)
            # Annotation
            elif after_unit_declaration and c == pu.ANNOTATION_OPEN_SYM:
                current_token = store_current_token()
                tokens.append(c)
                inside_annotation = True
                after_unit_declaration = False
            elif inside_annotation and c == pu.ANNOTATION_CLOSE_SYM:
                current_token = store_current_token()
                tokens.append(c)
                inside_annotation = False
            elif inside_annotation and c == pu.ANNOTATION_ASSIGNMENT_SYM:
                current_token = store_current_token()
                tokens.append(c)
            elif inside_annotation and c == pu.ANNOTATION_SEP:
                current_token = store_current_token()
                tokens.append(c)
            # Spaces
            elif c.isspace():
                if not current_token.isspace():
                    store_current_token()
                    current_token = c
            # Other
            else:
                if current_token.isspace() and not c.isspace():
                    current_token = store_current_token()
                current_token += c
                if after_unit_declaration:
                    after_unit_declaration = False
        store_current_token()

        if nb_closing_brackets_expected > 0:
            self.syntax_error("Line ends with open unit(s).", text, i)
        if inside_annotation:
            self.syntax_error("Line ends with an open annotation.", text, i)
        if inside_choice:
            self.syntax_error("Line ends with open choice(s).", text, i)
        if next_char_escaped:
            self.syntax_error("Line ends with unexpected escapement '"+
                              pu.ESCAPE_SYM+"'.", text, i)

        return tokens



    @staticmethod
    def _get_indentation(text):
        """
        Returns a string that is made
        of all the spaces at the beginning of `text`.
        """
        i = 0
        length = len(text)
        indentation = ""
        while i < length and text[i].isspace():
            indentation += text[i]
            i += 1
        if indentation != "":
            return indentation
        return None



"""
Module `chatette.cli.interactive_commands.unhide_command`.
Contains the strategy class that represents the interacive mode command
`unhide` which restores a unit definition that has been hidden.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.cli.interactive_commands.hide_command import HideCommand


class UnhideCommand(CommandStrategy):

    def execute(self, facade):
        """
        Implements the command `unhide` which restores a unit definition that
        was hidden from the parser of `facade`.
        """
        if len(self.command_tokens) < 3:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         'unhide <unit-type> "<unit-name>"')
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log("Unknown unit type: '" +
                                         str(self.command_tokens[1]) + "'.")
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])
        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(self.command_tokens[2])
            except SyntaxError:
                self.print_wrapper.error_log("Unit identifier couldn't be " + \
                                             "interpreted. Did you mean to " + \
                                             "escape some hashtags '#'?")
                return
            self.execute_on_unit(facade, unit_type, unit_name, variation_name)
        else:
            unit_names = [unit_name
                          for unit_name in HideCommand.stored_units[unit_type.name]
                          if unit_regex.match(unit_name)]
            if len(unit_names) == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")

            for unit_name in unit_names:
                self.execute_on_unit(facade, unit_type, unit_name)

    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        if variation_name is None:
            try:
                unit = HideCommand.stored_units[unit_type.name][unit_name]
                facade.parser.add_definition(unit_type, unit_name, unit)
                del HideCommand.stored_units[unit_type.name][unit_name]
                self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                        unit_name + "' was successfully restored.")
            except KeyError:
                self.print_wrapper.error_log(unit_type.name.capitalize() + " '" +
                                             unit_name + "' was not " +
                                             "previously hidden.")
            except ValueError:
                self.print_wrapper.error_log(unit_type.name.capitalize() + " '" +
                                             unit_name + "' is already defined " +
                                             "in the parser.")
        else:
            unit = None
            try:
                unit = facade.parser.get_definition(unit_name, unit_type)
            except KeyError:
                self.print_wrapper.error_log(unit_type.name.capitalize() + " '" + \
                                             unit_name + "' is not defined.")
                return
            try:
                rules = \
                    HideCommand.stored_variations[unit_type.name][unit_name][variation_name]
                if variation_name in unit.variations:
                    self.print_wrapper.error_log("Variation '" + variation_name + \
                                                 "' is already defined for " + \
                                                 unit_type.name + " '" + \
                                                 unit_name + "'.")
                    return
                unit.add_rules(rules, variation_name)
                self.print_wrapper.write("Variation '" + variation_name + \
                                         "' of " + unit_type.name + " '" + \
                                         unit_name + "' was successfully restored.")
            except KeyError:
                self.print_wrapper.error_log("Variation '" + variation_name + \
                                             "' of " + unit_type.name + " '" + \
                                             unit_name + "' was not " + \
                                             "previously hidden.")


    # Override abstract methods
    def finish_execution(self, facade):
        raise NotImplementedError()



"""
Module `chatette.modifiers.representation`
Contains structures that represent the possible modifiers
that could apply to unit declarations or sub-rules.
"""

class BaseModifiersRepr(object):
    def __init__(self, case_generation=False):
        # `case_generation` is a bool
        self.casegen = case_generation

class UnitDeclarationModifiersRepr(BaseModifiersRepr):
    def __init__(self, case_generation=False, variation_name=None,
                 argument_name=None):
        super(UnitDeclarationModifiersRepr, self).__init__(case_generation)
        self.variation_name = variation_name
        self.argument_name = argument_name

class WordGroupModifiersRepr(BaseModifiersRepr):
    def __init__(self, case_generation=False, randgen_name=None,
                 percentage_randgen=50):
        super(WordGroupModifiersRepr, self).__init__(case_generation)
        self.randgen_name = randgen_name
        self.percentage_randgen = percentage_randgen

class ChoiceModifiersRepr(BaseModifiersRepr):
    def __init__(self, case_generation=False, randgen=False):#,
                #  percentage_randgen=50):
        super(ChoiceModifiersRepr, self).__init__(case_generation)
        self.randgen = randgen
        # self.percentage_randgen = percentage_randgen

class ReferenceModifiersRepr(BaseModifiersRepr):
    def __init__(self, case_generation=False, randgen_name=None,
                 percentage_randgen=50, variation_name=None,
                 argument_value=None):
        super(ReferenceModifiersRepr, self).__init__(case_generation)
        self.randgen_name = randgen_name
        self.percentage_randgen = percentage_randgen
        self.variation_name = variation_name
        self.argument_value = argument_value



from chatette.units import UnitDefinition


class AliasDefinition(UnitDefinition):
    """
    This class represents the definition of an alias,
    containing all the rules that it can represent.
    """

    def __init__(self, name, modifiers, rules=None):
        super(AliasDefinition, self).__init__(name, modifiers, rules=rules)
        self.type = "alias"
    
    def _get_template_decl(self, variation=None):
        return '~' + super(AliasDefinition, self)._get_template_decl(variation)

    # Everything else is in the superclass



"""
Module `chatette.cli.interactive_commands.add_rule_command`.
Contains the strategy class that represents the interactive mode command
`add-rule` which allows to add a rule to a unit definition.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class AddRuleCommand(CommandStrategy):
    usage_str = 'add-rule <unit-type> "<unit-name>" "<rule>"'

    def execute(self, facade):
        # TODO support variations
        if len(self.command_tokens) < 4:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         self.usage_str)
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log("Unknown unit type: '" +
                                         str(self.command_tokens[1]) + "'.")
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])
        rule_str = CommandStrategy.remove_quotes(self.command_tokens[3])
        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(self.command_tokens[2])
            except SyntaxError:
                self.print_wrapper.error_log("Unit identifier couldn't be " + \
                                             "interpreted. Did you mean to " + \
                                             "escape some hashtags '#'?")
                return
            self._add_rule(facade.parser, unit_type, unit_name, variation_name,
                           rule_str)
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(facade.parser,
                                                          unit_type,
                                                          unit_regex):
                self._add_rule(facade.parser, unit_type, unit_name, None,
                               rule_str)
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")

    def _add_rule(self, parser, unit_type, unit_name, variation_name, rule_str):
        rule_tokens = parser.tokenizer.tokenize(rule_str)
        rule = parser.tokens_to_sub_rules(rule_tokens)

        unit = parser.get_definition(unit_name, unit_type)
        unit.add_rule(rule, variation_name)

        self.print_wrapper.write("Rule successfully added to " +
                                 unit_type.name + " '" + unit_name + "'.")


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



"""
Module `chatette.units.word`
Contains the rule contents characterizing words and word groups.
"""

from .group_rule_content import (GroupWordRuleContent, )
from .rule_content import (WordRuleContent, )



from __future__ import print_function

from chatette.units import Example, RuleContent, may_get_leading_space, \
                           randomly_change_case, with_leading_lower, with_leading_upper
from chatette.utils import choose
from chatette.parsing.parser_utils import add_escapement_back_in_choice_item, \
                                          CHOICE_SEP


class ChoiceRuleContent(RuleContent):
    """
    This class represents a choice as it can be contained in a rule,
    with its modifiers.
    Accepted modifiers:
        - leading-space: bool
        - casegen: bool
        - randgen: str
        # TODO: maybe more?
    """

    def __init__(self, name, leading_space=False, variation_name=None,
                 arg_value=None, casegen=False, randgen=None, percentage_gen=None,
                 parser=None):
        # NOTE: its name would be the unparsed text as a string
        if randgen is not None and randgen != "" and type(randgen) != bool:
            raise SyntaxError("Choices cannot have a named randgen, " +
                              "as was the case for '" + name + "': " + randgen +
                              " ('?' unescaped?)")
        if variation_name is not None:
            raise SyntaxError("Choices cannot have a variation, as was the " +
                              "case for '" + name + "' ('#' unescaped?)")
                              # TODO: change the symbol to a variable from parser_utils
        if arg_value is not None:
            raise SyntaxError("Choices cannot have an argument, as was the " +
                              "case for '" + name + "' ('$' unescaped?)")
        if percentage_gen is not None:
            raise SyntaxError("Choices cannot have an percentage for random " +
                              "generation, as was the case for '" + name + "'.")

        if randgen is None:
            randgen = False

        super(ChoiceRuleContent, self).__init__(
            name,
            leading_space=leading_space,
            variation_name=None,
            arg_value=None,
            casegen=casegen,
            randgen=randgen,
            percentage_gen=None,
            parser=None)

        self.choices = []
        self.casegen_checked = False


    def can_have_casegen(self):
        for choice in self.choices:
            if len(choice) > 0 and choice[0].can_have_casegen():
                return True
        return False

    def check_casegen(self):
        """Checks that casegen is applicable (at generation time)."""
        if not self.casegen_checked and self.casegen:
            if not self.can_have_casegen():
                self.casegen = False
            self.casegen_checked = True

    def get_max_nb_generated_examples(self):
        nb_possible_ex = 0
        for choice in self.choices:
            choice_nb_ex = 0
            for token in choice:
                current_nb_ex = token.get_max_nb_generated_examples()
                if choice_nb_ex == 0:
                    choice_nb_ex = current_nb_ex
                else:
                    choice_nb_ex *= current_nb_ex
            nb_possible_ex += choice_nb_ex

        if self.casegen:
            nb_possible_ex *= 2
        if self.randgen:
            nb_possible_ex += 1
        return nb_possible_ex


    def add_choice(self, choice):
        # ([RuleContent]) -> ()
        if len(choice) <= 0:
            return
        self.choices.append(choice)

    def add_choices(self, choices):
        # ([[RuleContent]]) -> ()
        interesting_choices = [choice for choice in choices if len(choice) > 0]
        if len(interesting_choices) <= 0:
            return
        self.choices.extend(interesting_choices)


    def generate_random(self, generated_randgens=None):
        if generated_randgens is None:
            generated_randgens = dict()

        self.check_casegen()

        # Manage randgen
        if self.randgen:
            return Example()

        if len(self.choices) <= 0:
            return Example()

        choice = choose(self.choices)
        generated_example = Example()

        for token in choice:
            generated_token = token.generate_random(generated_randgens)
            generated_example.text += generated_token.text
            generated_example.entities.extend(generated_token.entities)

        if self.casegen:
            generated_example.text = randomly_change_case(generated_example.text)
        if self.leading_space and may_get_leading_space(generated_example.text):
            generated_example.text = ' ' + generated_example.text

        return generated_example

    def generate_all(self):
        self.check_casegen()

        generated_examples = []
        if self.randgen:
            generated_examples.append(Example())

        for choice in self.choices:
            current_examples = []
            for token in choice:
                current_token_all_generations = token.generate_all()
                if len(current_examples) <= 0:
                    current_examples = [gen for gen in current_token_all_generations]
                else:
                    current_examples = [
                        Example(
                            partial_example.text + gen.text,
                            partial_example.entities + gen.entities
                        )
                        for partial_example in current_examples
                        for gen in current_token_all_generations]
            generated_examples.extend(current_examples)

        if self.leading_space:
            for (i, ex) in enumerate(generated_examples):
                if may_get_leading_space(ex.text):
                    generated_examples[i].text = ' ' + ex.text

        if self.casegen:
            tmp_buffer = []
            for ex in generated_examples:
                tmp_buffer.append(Example(with_leading_lower(ex.text), ex.entities))
                tmp_buffer.append(Example(with_leading_upper(ex.text), ex.entities))

        return generated_examples


    def print_DBG(self, nb_indent=0):
        indentation = nb_indent * '\t'
        print(indentation + self.name)
        print(indentation + "\tvariation name: " + str(self.variation_name))
        print(indentation + "\targ value: " + str(self.arg_value))
        print(indentation + "\tcasegen: " + str(self.casegen))
        print(indentation + "\trandgen: " + str(self.randgen) + " with percentage: "
              + str(self.percentgen))

        for choice in self.choices:
            print(indentation + "\tChoice:")
            for token in choice:
                token.print_DBG(nb_indent + 2)

    def as_string(self):
        """
        Returns the representation of the rule
        as it would be written in a template file.
        """
        result = ""
        for choice in self.choices:
            if result != "":
                result += CHOICE_SEP
            for sub_rule in choice:
                result += \
                    add_escapement_back_in_choice_item(sub_rule.as_string())
        if self.casegen:
            result = '&'+result
        if self.variation_name is not None:
            result += '#'+self.variation_name
        if self.randgen:
            result += '?'
            if self.percentgen != 50:
                result += '/'+str(self.percentgen)
        if self.arg_value is not None:
            result += '$'+self.arg_value
        result = '{' + result + '}'
        if self.leading_space:
            result = ' '+result
        return result



"""
Module `chatette.cli.interactive_commands.parse_command`.
Contains the strategy class that represents the interactive mode command
`parse` which parses a new template file.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class ParseCommand(CommandStrategy):

    def execute(self, facade):
        """
        Implements the command `parse`,
        parsing a new template file using the current parser.
        """
        if len(self.command_tokens) <= 1:
            self.print_wrapper.error_log("Missing template file path\nUsage: " +
                                         "'parse <filepath>'")
            return
        filepath = self.command_tokens[1]
        facade.parse_file(filepath)


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# TODO shouldn't generate twice the same statement

from chatette.utils import print_DBG, remove_duplicates


class Generator(object):
    """
    Using the info parsed from the input file, this class will generate
    a Rasa NLU dataset and dump it in a JSON file.
    If there were inconsistencies in the input file, they are likely to be
    detected here.
    """
    DEFAULT_MAX_NB_EXAMPLES_PER_INTENT = 1000000  # TODO: this might not be relevant anymore

    def __init__(self, parser):
        self.parser = parser
        self.max_nb_single_intent_examples = \
            Generator.DEFAULT_MAX_NB_EXAMPLES_PER_INTENT

    def set_max_nb_single_intent_examples(self, new_max):
        self.max_nb_single_intent_examples = new_max

    def generate_train(self):
        print_DBG("Generating training examples...")
        for intent_name in self.parser.intent_definitions:
            intent = self.parser.intent_definitions[intent_name]
            examples = intent.generate(self.max_nb_single_intent_examples)
            for example in examples:
                yield example

    def generate_test(self, training_examples=None):
        should_generate_test_set = False

        for intent_name in self.parser.intent_definitions:
            if self.parser.intent_definitions[intent_name].nb_testing_examples_asked is not None:
                should_generate_test_set = True
                break

        if should_generate_test_set:
            print_DBG("Generating testing examples...")
            for intent_name in self.parser.intent_definitions:
                intent = self.parser.intent_definitions[intent_name]
                examples = intent.generate(self.max_nb_single_intent_examples, training_examples)
                for example in examples:
                    yield example

    def get_entities_synonyms(self):
        """
        Makes a dict of all the synonyms of entities
        based on the slot value they are assigned.
        """
        synonyms = dict()
        for slot_definition in self.parser.slot_definitions:
            current_synonyms_dict = \
                self.parser.slot_definitions[slot_definition].get_synonyms_dict()
            for slot_value in current_synonyms_dict:
                if slot_value not in synonyms:
                    synonyms[slot_value] = current_synonyms_dict[slot_value]
                else:
                    synonyms[slot_value].extend(current_synonyms_dict[slot_value])
        return remove_duplicates(synonyms)


if __name__ == "__main__":
    # pylint: disable=wrong-import-position
    # pylint: disable=wrong-import-order
    import warnings

    warnings.warn("You are running the wrong file ('generator.py')." +
                  "The file that should be run is '__main__.py'.")



"""
Module `chatette.cli.interactive_commands.stats_command`.
Contains the strategy class that represents the interactive mode command
`stats` which shows statistics about the parsing.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class StatsCommand(CommandStrategy):

    def execute(self, facade):
        """Implements the command `stats`, printing parsing statistics."""
        self.print_wrapper.write("Statistics:")
        stats = facade.get_stats_as_str()
        self.print_wrapper.write(stats)


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



"""
Module `chatette.units`
Contains the definition of the units (i.e. rules and their contents)
that make up the Abstract Syntax Tree, and the logic associated to
their generation.
"""

from __future__ import print_function
import re
import sys
import random
from copy import deepcopy

from chatette.utils import choose
from chatette.modifiers.representation import UnitDeclarationModifiersRepr
from chatette.parsing.parser_utils import add_escapement_back_in_sub_rule


ENTITY_MARKER = "<<CHATETTE_ENTITY>>"
# HACK: prepends every entity (removed on output file writing)
#       to find back its index unambiguously.


class Example(object):
    """
    Represents an utterance (i.e. an example of sentence)
    that will later on be written in the output file(s).
    """
    def __init__(self, text=None, entities=None):# -> None:
        super(Example, self).__init__()

        if entities is None:
            entities = []
        if text is None:
            text = ""

        self.text = text
        self.entities = entities

    def __repr__(self):
        return "<'"+self.text+"' "+str(self.entities)+'>'
    def __str__(self):
        return self.text + '\n\tEntities: ' + str(self.entities)

    def __hash__(self):
        return hash(self.text+str(self.entities))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    def __ne__(self, other):
        return not self.__eq__(other)


def may_change_leading_case(text):
    """
    Checks whether the string `text` can
    change the letter case of its leading letter.
    """
    for c in text:
        if c.isalpha():
            return True
        if c.isspace():
            continue
        return False
    return False


def randomly_change_case(text):
    """
    Randomly set the case of the first letter of `text`.
    NOTE: this doesn't use `capitalize()` since we need to support changing the
          case of text that is already capitalized or indented.
    """
    if random.randint(0, 99) >= 50:
        return with_leading_lower(text)
    else:
        return with_leading_upper(text)


def with_leading_upper(text):
    """Returns `text` with a leading uppercase letter."""
    for (i, c) in enumerate(text):
        if not c.isspace():
            return text[:i] + text[i].upper() + text[(i + 1):]
    return text


def with_leading_lower(text):
    """Returns `text` with a leading lowercase letter."""
    for (i, c) in enumerate(text):
        if not c.isspace():
            return text[:i] + text[i].lower() + text[(i + 1):]
    return text


def may_get_leading_space(text):
    return text != "" and not text.startswith(' ')  # TODO: Add '\t'?


class UnitDefinition(object):
    """Superclass representing a unit definition."""

    def __init__(self, name, modifiers, rules=None):
        self.type = "unit"

        self.name = name
        if rules is None:
            self.rules = []
        else:
            self.rules = rules  # list of list of `RulesContent`s => [[RulesContent]]

        if modifiers is None:
            modifiers = UnitDeclarationModifiersRepr()
        self.modifiers = modifiers

        self.set_arg(modifiers.argument_name)

        self.variations = dict()
    @classmethod
    def from_raw_var(cls, name, rules=None, arg=None, casegen=False):
        modifiers = UnitDeclarationModifiersRepr(case_generation=casegen,
                                                 argument_name=arg)
        return cls(name, modifiers, rules)

    def set_arg(self, argument_identifier):
        self.argument_identifier = argument_identifier
        if argument_identifier is not None:
            PATTERN_ARG = r"(?<!\\)\$" + argument_identifier
            self.arg_regex = re.compile(PATTERN_ARG)
        else:
            self.arg_regex = None

    def __repr__(self):
        result = self.type+":"+self.name
        if self.modifiers.casegen:
            result = '&'+result
        return '<'+result+'>'

    def can_have_casegen(self):  # TODO: manage variations
        """
        Returns `True` if casegen may have an influence on
        any of the rules of this definition.
        """
        for rule in self.rules:
            if len(rule) > 0 and rule[0].can_have_casegen():
                return True
        return False

    def get_nb_rules(self, variation_name=None):
        """Returns the number of rules this definition contains."""
        if variation_name is None:
            return len(self.rules)
        return len(self.variations[variation_name])

    def get_max_nb_generated_examples(self, variation_name=None):
        """Returns the number of examples that can be generated by this token."""
        # TODO cache this
        relevant_rules = self.rules
        if variation_name is not None:
            if variation_name in self.variations:
                relevant_rules = self.variations[variation_name]
            else:
                raise SyntaxError("Couldn't find variation '" + variation_name +
                                  "' for " + self.type + " '" + self.name + "'")

        nb_possible_ex = 0
        for rule in relevant_rules:
            rule_nb_ex = 0
            for sub_unit_rule in rule:
                current_nb_ex = sub_unit_rule.get_max_nb_generated_examples()
                if current_nb_ex is None:
                    continue
                if rule_nb_ex == 0:
                    rule_nb_ex = current_nb_ex
                else:
                    rule_nb_ex *= current_nb_ex
            nb_possible_ex += rule_nb_ex

        if self.modifiers.casegen:
            nb_possible_ex *= 2
        return nb_possible_ex


    def add_rule(self, rule, variation_name=None):
        # (RuleContent, str) -> ()
        self.rules.append(rule)
        if variation_name is not None:
            if variation_name == "":
                raise SyntaxError("Defining a " + self.type + " with an empty name" +
                                  "is not allowed")
            if variation_name not in self.variations:
                self.variations[variation_name] = [rule]
            else:
                self.variations[variation_name].append(rule)

    def add_rules(self, rules, variation_name=None):
        # ([RuleContent], str) -> ()
        self.rules.extend(rules)
        if variation_name is not None:
            if variation_name == "":
                raise SyntaxError("Defining a " + self.type + " with an empty name" +
                                  "is not allowed")
            if variation_name not in self.variations:
                self.variations[variation_name] = rules
            else:
                self.variations[variation_name].extend(rules)

    def delete_variation(self, variation_name):
        """Deletes all rules that are related to a variation."""
        variation_rules = self.variations[variation_name]
        for rule in variation_rules:
            self.rules.remove(rule)
        del self.variations[variation_name]


    def generate_nb_examples(self, nb_examples, variation_name=None,
                             arg_value=None):
        """
        Generates `nb_examples` examples (at random) and returns a list of
        `Example`s containing the generated strings and entities inside it.
        Calls `generate_random` or `generate_all` internally.
        """
        max_nb_examples = self.get_max_nb_generated_examples(variation_name)
        if nb_examples is None or nb_examples >= max_nb_examples:
            return self.generate_all(variation_name, arg_value)
        if nb_examples > max_nb_examples/2:
            all_examples = self.generate_all(variation_name, arg_value)
            return random.sample(all_examples, nb_examples)
        examples = []
        for _ in range(nb_examples):
            new_example = self.generate_random(variation_name, arg_value)
            nb_iterations = 0
            while new_example in examples and nb_iterations < 50: # 50 is completely arbitrary
                new_example = self.generate_random(variation_name, arg_value)
            examples.append(new_example)
        return examples

    def generate_random(self, variation_name=None, arg_value=None):
        """
        Generates one of your rule at random and
        returns the string generated and the entities inside it as an `Example`.
        """
        if variation_name is None:
            chosen_rule = choose(self.rules)
        else:
            if variation_name not in self.variations:
                raise SyntaxError("Couldn't find a variation named '" +
                                  variation_name + "' for " + self.type + " '" + self.name + "'")
            chosen_rule = choose(self.variations[variation_name])

        if chosen_rule is None:  # No rule
            return Example()

        example_text = ""
        example_entities = []
        generated_examples = dict()

        for token in chosen_rule:
            generated_token = token.generate_random(generated_examples)
            example_text += generated_token.text
            example_entities.extend(generated_token.entities)

        if self.modifiers.casegen and self.can_have_casegen:
            example_text = randomly_change_case(example_text)

        # Replace `arg` inside the generated sentence
        if arg_value is not None and self.modifiers.argument_name is not None:
            example_text = self.arg_regex.sub(arg_value, example_text)
            # pylint: disable=anomalous-backslash-in-string
            example_text = example_text.replace(r"\$", "$")

        return Example(example_text, example_entities)

    def generate_all(self, variation_name=None, arg_value=None):
        generated_examples = []

        relevant_rules = self.rules
        if variation_name is not None:
            if variation_name in self.variations:
                relevant_rules = self.variations[variation_name]
            else:
                raise SyntaxError("Couldn't find variation '" +
                                  str(variation_name) + "' for " + str(self.type) +
                                  " '" + str(self.name) + "'")

        if not relevant_rules:  # No rules
            if variation_name is None:
                raise SyntaxError("No rules could be found for "+self.type+" '"+
                                  self.name+"'")
            else:
                raise SyntaxError("No rules could be found for "+self.type+" '"+
                                  self.name+"' (variation: '"+variation_name+"'")

        for rule in relevant_rules:
            examples_from_current_rule = []
            for sub_unit_rule in rule:
                sub_unit_possibilities = \
                    sub_unit_rule.generate_all()
                if len(examples_from_current_rule) <= 0:
                    examples_from_current_rule = sub_unit_possibilities
                else:
                    tmp_buffer = []
                    for ex in examples_from_current_rule:
                        for possibility in sub_unit_possibilities:
                            tmp_buffer.append(
                                Example(
                                    ex.text + possibility.text,
                                    ex.entities + possibility.entities
                                )
                            )
                    examples_from_current_rule = tmp_buffer
            generated_examples.extend(examples_from_current_rule)

        # Replace `arg` inside generated sentences
        if arg_value is not None and self.modifiers.argument_name is not None:
            for (i, ex) in enumerate(generated_examples):
                ex.text = self.arg_regex.sub(arg_value, ex.text)
                # pylint: disable=anomalous-backslash-in-string
                generated_examples[i].text = ex.text.replace(r"\$", "$")

        # Apply casegen
        if self.modifiers.casegen and self.can_have_casegen():
            tmp_examples = []
            for ex in generated_examples:
                (lower_ex, upper_ex) = (deepcopy(ex), deepcopy(ex))
                lower_ex.text = with_leading_lower(lower_ex.text)
                upper_ex.text = with_leading_upper(upper_ex.text)
                if lower_ex != upper_ex:
                    tmp_examples.append(lower_ex)
                    tmp_examples.append(upper_ex)
                else:
                    tmp_examples.append(ex)
            generated_examples = tmp_examples

        return generated_examples


    def print_DBG(self):
        print("\t" + self.type + ": " + self.name)
        print("\t\targument: " + str(self.modifiers.argument_name))
        print("\t\tcasegen: " + str(self.modifiers.casegen))
        print("\t\trules:")
        for rule in self.rules:
            print("\t\t\trule:")
            for content in rule:
                content.print_DBG(4)
        for variation in self.variations:
            print("\t\tvariation: " + variation)
            for rule in self.variations[variation]:
                print("\t\t\trule:")
                for content in rule:
                    content.print_DBG(3)
        print("")

    def short_desc_str(self):
        """
        Returns a str representing a short description of this unit description.
        """
        desc = self.type+": '"+self.name+"'\nmodifiers:\n"
        if self.modifiers.casegen:
            desc += "\t- casegen\n"
        if self.argument_identifier is not None:
            desc += "\t- argument: "+str(self.argument_identifier)+'\n'
        if desc.endswith("modifiers:\n"):
            desc += "\tNone\n"
        desc += str(len(self.variations)) + " variations"
        if len(self.variations) > 0:
            desc += ':'
        for (i,variation_name) in enumerate(self.variations):
            desc += "\n\t- " + variation_name
            if i >= 12:
                desc += "\n\t- ..."
                break
        return desc

    def get_template_description(self):
        """
        Returns a string that should be equal to the definition of the unit
        in the template files (i.e. a template string on one or several lines).
        """
        result = self._get_template_decl() + '\n'
        rules_str = '\n'.join(['\t' + e for e in self._get_template_rules()]) + \
                    '\n'
        if not rules_str.isspace():
            result += rules_str
        else:
            result = ""
        for variation_name in self.variations:
            current_decl = self._get_template_decl(variation_name) + '\n'
            rules_str = '\n'.join(
                            ['\t' + e
                             for e in self._get_template_rules(variation_name)]
                        ) + '\n'
            if not rules_str.isspace():
                result += current_decl + rules_str
        return result
    def _get_template_decl(self, variation=None):
        """
        Returns the template string representing the declaration line of this
        unit definition (with potentially a variation).
        Subclasses should override this method.
        """
        # TODO use parser utils variables
        result = '['
        if self.modifiers.casegen:
            result += '&'
        result += self.name
        if variation is not None:
            if variation not in self.variations:
                raise ValueError("Tried to get template for a unit declaration " +
                                 "with an invalid variation name")
            result += '#' + variation
        if self.argument_identifier is not None:
            result += '$' + self.argument_identifier
        result += ']'
        return result
    def _get_template_rules(self, variation=None):
        """
        Returns a list of template string that represent each rule of this unit
        (potentially with a variation).
        This method may need to be overriden by subclasses.
        """
        rules = self.rules
        if variation is not None:
            rules = self.variations[variation]
        rule_templates = []
        for rule in rules:
            rule_templates.append(''.join([sub_rule.as_string()
                                           for sub_rule in rule]))
        return rule_templates


class RuleContent(object):
    """
    Superclass represents anything that can be inside a rule:
    for words and word groups, it generates as is;
    for units, it is a link to a definition that can be generated.
    The rule also contains modifier used during the generation, such as:
        - leading-space: bool (a leading space will be added to the generated str)
        - casegen: bool (the first letter may be in upper- or lowercase)
        - randgen: str (if not `None` it might not be generated,
                        and if it is the same string than for another rule,
                        it will generate iff the other rule generated (and vice-versa))
        - percentgen: int (if `randgen` is enable, this is the percentage of
                           of chances that this rule will generate something)
        - arg: str (represents the identifier of an argument inside the rule,
                    which will be replaced by a value given upon generation)
        - variation-name: str (identifies which variation of the definition
                               we are calling)
    """

    def __init__(self, name, leading_space=False, variation_name=None,
                 arg_value=None, casegen=False, randgen=None, percentage_gen=50,
                 parser=None):
        if name is None or name == "":
            raise SyntaxError("Tried to create content without a contents (or a name)")
        self.name = name
        self.variation_name = variation_name
        self.arg_value = arg_value

        self.leading_space = leading_space

        self.casegen = casegen
        self.randgen = randgen
        if percentage_gen is not None:
            self.percentgen = int(percentage_gen)
        else:
            self.percentgen = 50

        self.parser = parser

    def can_have_casegen(self):
        """Returns `True` if casegen can have an influence on this rule."""
        return False

    def get_max_nb_generated_examples(self):
        """Returns the number of examples that can be generated by this token."""
        pass


    def generate_random(self, generated_randgens=None):
        """
        Returns a string and its entities randomly generated from the rules the
        object represents. May return an empty string if `randgen` is enabled.
        `generated_randgens` is a dict of all the randgen names that have been
        decided as to generate or not, i.e. if "some randgen" is in
        `generated_randgens` and its value is `True`, all contents with this
        randgen must be generated; if its value is `False`, it cannot be
        generated; otherwise you can choose and put it into `generated_randgens`.
        """
        return Example()

    def generate_all(self):
        """
        Returns a list of all the strings and entities that can be generated
        from the rules this object represents. May include the empty string if
        it can be generated.
        """
        return [Example()]


    def print_DBG(self, nb_indent=0):
        # (int) -> ()
        indentation = nb_indent * '\t'
        print(indentation + self.name)
        print(indentation + "\tvariation name: " + str(self.variation_name))
        print(indentation + "\targ value: " + str(self.arg_value))
        print(indentation + "\tcasegen: " + str(self.casegen))
        print(indentation + "\trandgen: " + str(self.randgen) + " with percentage: "
              + str(self.percentgen))

    def __repr__(self):
        result = "'"+self.name+"'"
        if self.casegen:
            result = '&'+result
        if self.leading_space:
            result = ' '+result
        if self.variation_name is not None:
            result += '#'+self.variation_name
        if self.randgen is not None:
            result += '?'+str(self.randgen)
            if self.percentgen != 50:
                result += '/'+str(self.percentgen)
        if self.arg_value is not None:
            result += '$'+self.arg_value
        return '<'+result+'>'

    def as_string(self):
        """
        Returns the representation of the rule
        as it would be written in a template file.
        """
        result = add_escapement_back_in_sub_rule(self.name)
        if self.casegen:
            result = '&'+result
        if self.variation_name is not None:
            result += '#'+self.variation_name
        if self.randgen is not None:
            result += '?'+str(self.randgen)
            if self.percentgen != 50:
                result += '/'+str(self.percentgen)
        if self.arg_value is not None:
            result += '$'+self.arg_value
        if self.leading_space:
            result = ' '+result
        return result



from random import randint

from chatette.parsing.parser_utils import UnitType, \
                                          add_escapement_back_in_unit_ref
from chatette.units import Example, RuleContent, may_get_leading_space, \
                           may_change_leading_case, randomly_change_case, \
                           with_leading_lower, with_leading_upper


class AliasRuleContent(RuleContent):
    """
    This class represents an alias as it can be contained in a rule,
    with its modifiers.
    Accepted modifiers:
        - leading-space: bool
        - casegen: bool
        - randgen: str
        - percentgen: int
        - arg: str
        - variation-name: str
    """

    def __init__(self, name, leading_space=False, variation_name=None, arg_value=None,
                 casegen=False, randgen=None, percentage_gen=50, parser=None):
        # TODO replace parser by self definition
        super(AliasRuleContent, self).__init__(name, leading_space=leading_space,
                                               variation_name=variation_name,
                                               arg_value=arg_value, casegen=casegen,
                                               randgen=randgen, percentage_gen=percentage_gen,
                                               parser=parser)
        self.casegen_checked = False


    def get_max_nb_generated_examples(self):
        nb_possible_ex = self.parser.get_definition(self.name, UnitType.alias) \
            .get_max_nb_generated_examples(self.variation_name)
        if self.casegen:
            nb_possible_ex *= 2
        if self.randgen is not None:
            nb_possible_ex += 1
        return nb_possible_ex

    def can_have_casegen(self):
        return self.parser.get_definition(self.name, UnitType.alias) \
            .can_have_casegen()

    def check_casegen(self):
        """Checks that casegen is applicable (at generation time)."""
        if not self.casegen_checked and self.casegen:
            if not self.can_have_casegen():
                self.casegen = False
            self.casegen_checked = True


    def generate_random(self, generated_randgens=None):
        if generated_randgens is None:
            generated_randgens = dict()

        self.check_casegen()

        # Manage randgen
        if self.randgen is not None and self.randgen in generated_randgens:
            if generated_randgens[self.randgen]:
                pass  # Must be generated
            else:
                return Example()  # Cannot be generated
        elif self.randgen is not None:
            if randint(0, 99) >= self.percentgen:
                # Don't generated this randgen
                if self.randgen != "":
                    generated_randgens[self.randgen] = False
                return Example()
            elif self.randgen != "":
                # Generate this randgen
                generated_randgens[self.randgen] = True

        generated_example = self.parser.get_definition(self.name, UnitType.alias) \
            .generate_random(self.variation_name, self.arg_value)

        if self.casegen:
            generated_example.text = \
                randomly_change_case(generated_example.text)

        if self.leading_space and may_get_leading_space(generated_example.text):
            generated_example.text = ' ' + generated_example.text

        return generated_example

    def generate_all(self):
        self.check_casegen()

        generated_examples = []
        if self.randgen is not None:
            generated_examples.append(Example())

        aliases = self.parser \
                      .get_definition(self.name, UnitType.alias) \
                      .generate_all(self.variation_name, self.arg_value)

        generated_examples.extend(aliases)

        if self.leading_space:
            for (i, ex) in enumerate(generated_examples):
                if may_get_leading_space(ex.text):
                    generated_examples[i].text = ' ' + ex.text
        if self.casegen:
            tmp_buffer = []
            for ex in generated_examples:
                if may_change_leading_case(ex.text):
                    tmp_buffer.append(Example(with_leading_lower(ex.text), ex.entities))
                    tmp_buffer.append(Example(with_leading_upper(ex.text), ex.entities))
                else:
                    tmp_buffer.append(ex)
            generated_examples = tmp_buffer
        return generated_examples


    def as_string(self):
        """
        Returns the representation of the rule
        as it would be written in a template file.
        """
        result = add_escapement_back_in_unit_ref(self.name)
        if self.casegen:
            result = '&'+result
        if self.variation_name is not None:
            result += '#'+self.variation_name
        if self.randgen is not None:
            result += '?'+str(self.randgen)
            if self.percentgen != 50:
                result += '/'+str(self.percentgen)
        if self.arg_value is not None:
            result += '$'+self.arg_value
        result = "~[" + result + ']'
        if self.leading_space:
            result = ' '+result
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module `chatette.parsing.parser_utils`
Contains utility functions that are specific to
the parsing of template files.
"""


import re
from enum import Enum

from chatette import deprecations
import chatette.modifiers.representation as mods

COMMENT_SYM_DEPRECATED = ';'
COMMENT_MARKER = '//'
ESCAPE_SYM = '\\'

ALIAS_SYM = '~'
SLOT_SYM = '@'
INTENT_SYM = '%'
UNIT_OPEN_SYM = '['  # This shouldn't be changed
UNIT_CLOSE_SYM = ']'  # id.

ANNOTATION_OPEN_SYM = '('
ANNOTATION_CLOSE_SYM = ')'
ANNOTATION_SEP = ','
ANNOTATION_ASSIGNMENT_SYM = ':'
ANNOTATION_IGNORED_SYM = "'"

CHOICE_OPEN_SYM = r'{'
CHOICE_CLOSE_SYM = r'}'
CHOICE_SEP = '/'  # TODO: deprecate and rather use '|'

VARIATION_SYM = '#'
RAND_GEN_SYM = '?'  # This shouldn't be changed
PERCENT_GEN_SYM = '/'
CASE_GEN_SYM = '&'
ARG_SYM = '$'  # This shouldn't be changed

ALT_SLOT_VALUE_NAME_SYM = '='
ALT_SLOT_VALUE_FIRST_SYM = '/'

INCLUDE_FILE_SYM = '|'

# TODO add special characters at the beginning of those to prevent people from
#      using them by chance
RESERVED_VARIATION_NAMES = ["all-variations-aggregation", "rules",
                            "nb-gen-asked", "arg"]


PATTERN_COMMENT_DEPRECATED = re.compile(r"(?<!\\)" + COMMENT_SYM_DEPRECATED)
PATTERN_COMMENT = re.compile(r"(?<!\\)" + COMMENT_MARKER)

PATTERN_NB_TRAIN_EX_KEY = re.compile(r"'?train(ing)?'?")
PATTERN_NB_TEST_EX_KEY = re.compile(r"'?test(ing)?'?")


class UnitType(Enum):
    """Enumeration of all possible types of unit declarations."""
    alias = 1
    slot = 2
    intent = 3


class SubRuleType(Enum):  # TODO move this into unit defintions
    """
    Enumeration of all possible types of units.
    Note: word is not considered a 'special' unit (others are).
    """
    word = 1  # simple word, no other info needed
    word_group = 2  # word group with modifiers
    alias = 3  # alias with modifiers
    slot = 4  # slot with modifiers
    intent = 5  # intent with modifiers and generation number
    choice = 6  # choice with contained units


def strip_comments(text):
    """Returns the text without the comments (and right stripped)."""
    if text is None:
        return None
    elif text == "":
        return ""
    match = PATTERN_COMMENT.search(text)
    match_deprecated = PATTERN_COMMENT_DEPRECATED.search(text)
    if match_deprecated is not None:
        deprecations.warn_semicolon_comments()

    if match is None and match_deprecated is None:
        return text.rstrip()
    elif match_deprecated is None:
        return text[:match.start()].rstrip()
    elif match is None:
        return text[:match_deprecated.start()].rstrip()
    else:
        if match.start() <= match_deprecated.start():
            return text[:match.start()].rstrip()
        return text[:match_deprecated.start()].rstrip()

def remove_escapement(text):
    # pylint: disable=anomalous-backslash-in-string
    r"""
    Returns `text` were all escaped characters
    have been removed their escapement character (e.g. `\?` becomes `?`).
    Note that escaped dollar sign ($) are kept escaped until generation
    to avoid a possible bug with argument replacement.
    """
    if text is None:
        return None
    if ESCAPE_SYM not in text:
        return text
    # Note there might be better ways to do this with regexes
    # (but they have fixed-length negative lookback)
    result = ""
    escaped = False
    for c in text:
        if escaped and c == ARG_SYM:  # Keep \$ until generation
            result += ESCAPE_SYM + ARG_SYM
            escaped = False
        elif escaped:
            result += c
            escaped = False
        elif c == ESCAPE_SYM:
            escaped = True
        else:
            result += c
    return result

# NOTE: there should be a most pretty way to deal with escapement
#       (especially when putting it back into strings).

def add_escapement_back_for_not_comments(text):
    """
    Considering that `text` contains no comment,
    escape comment markers and returns the new text.
    This function is needed because comment markers are several characters long.
    @pre: there is no comments anymore in `text`.
    """
    return text.replace(COMMENT_MARKER, ESCAPE_SYM+COMMENT_MARKER)
def add_escapement_back_in_sub_rule(text):
    """
    Put back escapement where it belongs for sub-rules
    (words and word groups) where escapement have been removed.
    """
    # BUG $ are always escaped
    escaped_text = ""
    for c in text:
        if should_be_escaped_char(c):
            escaped_text += ESCAPE_SYM + c
        else:
            escaped_text += c
    return add_escapement_back_for_not_comments(escaped_text)
def add_escapement_back_in_word(text):
    """
    Put escapement back where it belongs in words
    where escapement has been removed.
    """
    escaped_text = ""
    for c in text:
        if c == ESCAPE_SYM or is_boundary_sym(c) or is_comment_sym(c):
            escaped_text += ESCAPE_SYM + c
        else:
            escaped_text += c
    escaped_text = add_escapement_back_for_not_comments(escaped_text)
    return escaped_text.replace(ESCAPE_SYM+ESCAPE_SYM+ARG_SYM, ESCAPE_SYM+ARG_SYM)
def add_escapement_back_in_group(text):
    """
    Put escapement back where it belongs in word groups
    where escapement has been removed.
    """
    escaped_text = ""
    for c in text:
        if (   c == ESCAPE_SYM or is_boundary_sym(c)
            or is_group_modifier_sym(c) or is_comment_sym(c)):
            escaped_text += ESCAPE_SYM + c
        else:
            escaped_text += c
    escaped_text = add_escapement_back_for_not_comments(escaped_text)
    return escaped_text.replace(ESCAPE_SYM+ESCAPE_SYM+ARG_SYM, ESCAPE_SYM+ARG_SYM)
def add_escapement_back_in_unit_ref(unit_name):
    """
    Put back escapement where it belongs for unit names in references
    where escapement have been removed.
    """
    escaped_text = ""
    for c in unit_name:
        if (   c == ESCAPE_SYM or is_boundary_sym(c) or is_comment_sym(c)
            or is_unit_ref_modifier_sym(c)):
            escaped_text += ESCAPE_SYM + c
        else:
            escaped_text += c
    return add_escapement_back_for_not_comments(escaped_text)
def add_escapement_back_in_choice_item(text):
    """
    Put escapement back where it belongs in choice items
    where escapement has been removed.
    Add escapement only in items where escapement has already been added,
    not knowing they are part of a choice.
    """
    escaped_text = ""
    escaped = False
    for c in text:
        if escaped:
            escaped_text += c
            escaped = False
            continue
        if c == ESCAPE_SYM:
            escaped_text += c
            escaped = True
            continue
        if is_choice_special_sym(c):
            escaped_text += ESCAPE_SYM + c
        else:
            escaped_text += c
    return escaped_text
def add_escapement_back_in_unit_decl(unit_name):
    """
    Put back escapement where it belongs for unit names in declarations
    where escapement have been removed.
    NOTE: not used at the moment
    """
    escaped_text = ""
    for c in unit_name:
        print(c)
        if (   c == ESCAPE_SYM or is_boundary_sym(c)
            or is_unit_decl_modifier_sym(c) or is_arg_sym(c) or is_comment_sym(c)):
            escaped_text += ESCAPE_SYM + c
        else:
            escaped_text += c
    return add_escapement_back_for_not_comments(escaped_text)


def should_be_escaped_char(text):
    """
    Returns `True` if `text` is 1 character that must be escaped in order to be
    considered part of a word.
    """
    return is_special_sym(text) or text == COMMENT_MARKER \
                                or text == COMMENT_SYM_DEPRECATED \
                                or text == ALT_SLOT_VALUE_NAME_SYM

def is_special_sym(text):
    """Returns `True` if `text` is a string made of only 1 special character."""
    return text == ALIAS_SYM or text == SLOT_SYM or text == INTENT_SYM or \
           text == UNIT_OPEN_SYM or text == UNIT_CLOSE_SYM or \
           text == VARIATION_SYM or text == RAND_GEN_SYM or \
           text == PERCENT_GEN_SYM or text == CASE_GEN_SYM or \
           text == ARG_SYM or text == CHOICE_OPEN_SYM or \
           text == CHOICE_CLOSE_SYM
def is_comment_sym(text):
    """Returns `True` iff `text` is a symbol introducing a comment."""
    return text in (COMMENT_MARKER, COMMENT_SYM_DEPRECATED)
def is_boundary_sym(text):
    """
    Returns `True` iff `text` is a symbol
    that makes the boundaries of special sub-rules.
    """
    return text in (ALIAS_SYM, SLOT_SYM, INTENT_SYM, UNIT_OPEN_SYM,
                    UNIT_CLOSE_SYM, CHOICE_OPEN_SYM, CHOICE_CLOSE_SYM)
def is_arg_sym(text):
    """Returns `True` iff `text` is the special symbol introducing arguments."""
    return text == ARG_SYM
def is_group_modifier_sym(text):
    """
    Returns `True` iff `text` is a special symbol introducing a modifier that
    can label word groups.
    """
    return text in (CASE_GEN_SYM, RAND_GEN_SYM, PERCENT_GEN_SYM)
def is_unit_ref_modifier_sym(text):
    """
    Returns `True` iff ´text´ is a special symbol introducing a modifier that
    can label unit references.
    """
    return text in (CASE_GEN_SYM, RAND_GEN_SYM, PERCENT_GEN_SYM, VARIATION_SYM,
                    ARG_SYM)
def is_unit_decl_modifier_sym(text):
    """
    Returns `True` iff `text` is a special symbol introducing a modifier that
    can label unit declarations.
    """
    return text in (CASE_GEN_SYM, VARIATION_SYM, ARG_SYM)
def is_choice_special_sym(text):
    """
    Returns `True` iff `text` is a special symbol in a choice (and in nothing
    else). This doesn't take into account boundary symbols.
    """
    return text == CHOICE_SEP

def is_unit_type_sym(text):
    """Returns `True` if `text` is a unit special symbol (`~`, `@` or `%`)."""
    return text == ALIAS_SYM or text == SLOT_SYM or text == INTENT_SYM

def is_start_unit_sym(char):
    """Checks if character `char` is the starting character of a special unit."""
    return char == UNIT_OPEN_SYM or char == ALIAS_SYM or \
           char == SLOT_SYM or char == INTENT_SYM


def get_unit_type_from_sym(sym):
    """
    Returns which unit type corresponds to this special character.
    Returns `None` if it doesn't correspond to any unit type.
    """
    if sym == ALIAS_SYM:
        return UnitType.alias
    if sym == SLOT_SYM:
        return UnitType.slot
    if sym == INTENT_SYM:
        return UnitType.intent
    return None

def get_declaration_interior(tokens):
    """
    Returns a list of tokens that represent the inside of the declaration
    that is initiated on this line.
    Returns `None` if there is no unit declared in `tokens`.
    """
    length = len(tokens)
    starting_index = 0
    while starting_index < length and tokens[starting_index] != UNIT_OPEN_SYM:
        starting_index += 1
    starting_index += 1
    if starting_index >= length:
        return None

    end_index = starting_index
    nb_closing_brackets_expected = 1
    while end_index < length and nb_closing_brackets_expected > 0:
        if tokens[end_index] == UNIT_OPEN_SYM:
            nb_closing_brackets_expected += 1
        elif tokens[end_index] == UNIT_CLOSE_SYM:
            nb_closing_brackets_expected -= 1
        end_index += 1
    end_index -= 1
    if end_index == starting_index:
        return None

    return tokens[starting_index:end_index]

def get_annotation_interior(tokens):
    """
    Returns a list of tokens that represent the inside of the annotation
    that is present on this line.
    Returns `None` if there is no annotation in `tokens`.
    """
    length = len(tokens)
    starting_index = 0
    while starting_index < length and tokens[starting_index] != ANNOTATION_OPEN_SYM:
        starting_index += 1
    starting_index += 1
    if starting_index >= length:
        return None

    end_index = starting_index
    nb_closing_brackets_expected = 1
    while end_index < length and nb_closing_brackets_expected > 0:
        if tokens[end_index] == ANNOTATION_OPEN_SYM:
            nb_closing_brackets_expected += 1
        elif tokens[end_index] == ANNOTATION_CLOSE_SYM:
            nb_closing_brackets_expected -= 1
        end_index += 1
    end_index -= 1
    if end_index == starting_index:
        return None

    return tokens[starting_index:end_index]


def check_declaration_validity(tokens_unit_inside):
    """
    Check that the interior of a declaration is syntactically legal.
    Raises a `SyntaxError` if the declaration is invalid.
    The constraints checked are:
    - there is only one modifier of each type
    - there are no randgen or percentgen modifiers
    - `&` is at the beginning of the declaration (or nowhere)
    - there is a name after `#`
    - there is a value after `$`
    - there is a name either after `&` or at the beginning
    - the variation names are not reserved
    """
    casegen_count = tokens_unit_inside.count(CASE_GEN_SYM)
    if casegen_count > 1:
        raise SyntaxError("There can be only one case generation modifier "+
                          "in a unit declaration.")
    if casegen_count == 1 and tokens_unit_inside.index(CASE_GEN_SYM) != 0:
        raise SyntaxError("Case generation modifiers have to be at the start "+
                          "of a unit declaration.")

    if casegen_count == 0 and is_special_sym(tokens_unit_inside[0]):
        raise SyntaxError("Unit declarations must be named.")
    elif casegen_count == 1 and len(tokens_unit_inside) <= 1:
        raise SyntaxError("Unit declarations must be named.")
    elif casegen_count == 1 and is_special_sym(tokens_unit_inside[1]):
        raise SyntaxError("Unit declarations must be named.")

    variation_count = tokens_unit_inside.count(VARIATION_SYM)
    if variation_count > 1:
        raise SyntaxError("There can be only one variation modifier "+
                          "in a unit declaration.")
    if variation_count == 1:
        variation_name_index = tokens_unit_inside.index(VARIATION_SYM)+1
        if     variation_name_index >= len(tokens_unit_inside) \
            or is_special_sym(tokens_unit_inside[variation_name_index]):
            raise SyntaxError("Variations must be named.")
        variation_name = tokens_unit_inside[variation_name_index]
        if variation_name in RESERVED_VARIATION_NAMES:
            raise SyntaxError("The following variation names are reserved: "+
                              str(RESERVED_VARIATION_NAMES)+". Please don't "+
                              "use them.")

    argument_count = tokens_unit_inside.count(ARG_SYM)
    if argument_count > 1:
        raise SyntaxError("There can be only one argument modifier "+
                          "per unit declaration.")
    if argument_count == 1:
        argument_name_index = tokens_unit_inside.index(ARG_SYM)+1
        if     argument_name_index >= len(tokens_unit_inside) \
            or is_special_sym(tokens_unit_inside[argument_name_index]):
            raise SyntaxError("Arguments must be named.")

    # TODO remove the following because you should allow ? and / in declarations?
    #      or the tokenizer should not consider them special characters in this
    #      case
    randgen_count = tokens_unit_inside.count(RAND_GEN_SYM)
    if randgen_count > 0:
        raise SyntaxError("Unit declarations cannot take a random generation "+
                          "modifier.")
    percentgen_count = tokens_unit_inside.count(PERCENT_GEN_SYM)
    if percentgen_count > 0:
        raise SyntaxError("Unit declarations cannot take a percentage for "+
                          "the random generation modifier.")


def check_reference_validity(tokens_unit_inside):
    """
    Check that the interior of a reference is syntactically legal.
    Raises a `SyntaxError` if the reference is invalid.
    The constraints checked are:
    - there is only one modifier of each type
    - `/` is not there unless `?` is there
    - there is a number between 0 and 100 if `/` is present
    - `&` is at the beginning of the declaration (or nowhere)
    - there is a name after `#`
    - there is a name either after `&` or at the beginning
    """
    casegen_count = tokens_unit_inside.count(CASE_GEN_SYM)
    if casegen_count > 1:
        raise SyntaxError("There can be only one case generation modifier "+
                          "in a unit reference.")
    if casegen_count == 1 and tokens_unit_inside.index(CASE_GEN_SYM) != 0:
        raise SyntaxError("Case generation modifiers have to be at the start "+
                          "of a unit reference.")

    if casegen_count == 0 and is_special_sym(tokens_unit_inside[0]):
        raise SyntaxError("Unit references must be named.")
    elif casegen_count == 1 and len(tokens_unit_inside) <= 1:
        raise SyntaxError("Unit references must be named.")
    elif casegen_count == 1 and is_special_sym(tokens_unit_inside[1]):
        raise SyntaxError("Unit references must be named.")

    variation_count = tokens_unit_inside.count(VARIATION_SYM)
    if variation_count > 1:
        raise SyntaxError("There can be only one variation modifier "+
                          "in a unit reference.")
    if variation_count == 1:
        variation_name_index = tokens_unit_inside.index(VARIATION_SYM)+1
        if     variation_name_index >= len(tokens_unit_inside) \
            or is_special_sym(tokens_unit_inside[variation_name_index]):
            raise SyntaxError("Variations must be named.")
        variation_name = tokens_unit_inside[variation_name_index]
        if variation_name in RESERVED_VARIATION_NAMES:
            raise SyntaxError("The following variation names are reserved: "+
                              str(RESERVED_VARIATION_NAMES)+". Please don't "+
                              "use them.")

    argument_count = tokens_unit_inside.count(ARG_SYM)
    if argument_count > 1:
        raise SyntaxError("There can be only one argument modifier "+
                          "per unit reference.")
    # if argument_count == 1:
    #     argument_name_index = tokens_unit_inside.index(ARG_SYM)+1
    #     if     argument_name_index >= len(tokens_unit_inside) \
    #         or is_special_sym(tokens_unit_inside[argument_name_index]):
    #         raise SyntaxError("Arguments must be named.")

    randgen_count = tokens_unit_inside.count(RAND_GEN_SYM)
    if randgen_count > 1:
        raise SyntaxError("There can be only one random generation modifier "+
                          "per unit reference.")
    percentgen_count = tokens_unit_inside.count(PERCENT_GEN_SYM)
    if percentgen_count > 1:
        raise SyntaxError("There can be only one percentage for generation "+
                          "modifier per unit reference.")
    if percentgen_count == 1 and randgen_count == 0:
        raise SyntaxError("There cannot be a percentage for generation "+
                          "modifier if there is no random generation modifier "+
                          "(did you mean to escape '"+PERCENT_GEN_SYM+"'?)")
    if percentgen_count == 1:
        index_randgen = tokens_unit_inside.index(RAND_GEN_SYM)
        index_percentgen = tokens_unit_inside.index(PERCENT_GEN_SYM)
        if index_randgen > index_percentgen:
            raise SyntaxError("A percentage for generation modifier must "+
                              "always be right after the random generation "+
                              "modifier.")
        if index_percentgen == len(tokens_unit_inside)-1:
            raise SyntaxError("No percentage found after the special symbol "+
                              "for percentage modifier.")
        try:
            percentgen = int(tokens_unit_inside[index_percentgen+1])
        except ValueError:
            raise SyntaxError("Percentage for generation modifiers need to be "+
                              "an integer.")
        if percentgen < 0 or percentgen > 100:
            raise SyntaxError("Percentage for generation modifiers need to be "+
                              "between 0 and 100.")

def check_choice_validity(tokens_choice_inside):
    """
    Check that the interior of a choice is syntactically legal.
    Deals with word groups as well.
    Raises a `SyntaxError` if the choice is invalid.
    As any sub-rules can be inside choices, we cannot check anything except
    that the last tokens is not a separator (or the two-to-last one if the
    last one is a random generation modifier).
    """
    # TODO: deprecate `/` as choice separators AND percentgen
    # percentgen_count = tokens_choice_inside.count(PERCENT_GEN_SYM)
    # if percentgen_count > 0:
    #     raise SyntaxError("Choices cannot take a percentage for generation "+
    #                       "modifier.")
    if len(tokens_choice_inside) > 0:
        if tokens_choice_inside[-1] == CHOICE_SEP:
            raise SyntaxError("Choice cannot end with a choice separator. " +
                              "Did you forget to escape the last character?")
        if (    len(tokens_choice_inside) > 1
            and tokens_choice_inside[-1] == RAND_GEN_SYM
            and tokens_choice_inside[-2] == CHOICE_SEP):
            raise SyntaxError("Choice ends with an empty choice item. " +
                              "Did you forget to escape the choice separator?")


def check_word_group_validity(tokens_word_group_inside):
    """
    Check that the interior of a choice is syntactically legal.
    Deals with word groups as well.
    Raises a `SyntaxError` if the choice is invalid.
    The constraints checked are:
    - there is only one modifier of each type
    - `/` and `#` are not there
    - `&` is at the beginning of the declaration (or nowhere)
    - choices are separated by '/' (not checked as there can be 0 or 1 choice)
    """
    casegen_count = tokens_word_group_inside.count(CASE_GEN_SYM)
    if casegen_count > 1:
        raise SyntaxError("There can be only one case generation modifier "+
                          "in a word group.")
    if casegen_count == 1 and tokens_word_group_inside.index(CASE_GEN_SYM) != 0:
        raise SyntaxError("Case generation modifiers have to be at the start "+
                          "of a word group.")

    variation_count = tokens_word_group_inside.count(VARIATION_SYM)
    if variation_count > 0:
        raise SyntaxError("Word groups cannot take variation modifiers.")

    argument_count = tokens_word_group_inside.count(ARG_SYM)
    if argument_count > 0:
        raise SyntaxError("Word groups cannot take arguments.")

    randgen_count = tokens_word_group_inside.count(RAND_GEN_SYM)
    if randgen_count > 1:
        raise SyntaxError("There can be only one random generation modifier "+
                          "per word group.")
    percentgen_count = tokens_word_group_inside.count(PERCENT_GEN_SYM)
    if percentgen_count > 1:
        raise SyntaxError("There can be only one percentage for generation "+
                          "modifier per word group.")
    if percentgen_count == 1 and randgen_count == 0:
        raise SyntaxError("There cannot be a percentage for generation "+
                          "modifier if there is no random generation modifier "+
                          "(did you mean to escape '"+PERCENT_GEN_SYM+"'?)")
    if percentgen_count == 1:
        index_randgen = tokens_word_group_inside.index(RAND_GEN_SYM)
        index_percentgen = tokens_word_group_inside.index(PERCENT_GEN_SYM)
        if index_randgen > index_percentgen:
            raise SyntaxError("A percentage for generation modifier must "+
                              "always be right after the random generation "+
                              "modifier.")
        if index_percentgen == len(tokens_word_group_inside)-1:
            raise SyntaxError("No percentage found after the special symbol "+
                              "for percentage modifier.")
        try:
            percentgen = int(tokens_word_group_inside[index_percentgen+1])
        except ValueError:
            raise SyntaxError("Percentage for generation modifiers need to be "+
                              "an integer.")
        if percentgen < 0 or percentgen > 100:
            raise SyntaxError("Percentage for generation modifiers need to be "+
                              "between 0 and 100.")


def find_name(tokens_inside_unit):
    """
    Finds the name of the unit from the tokens that represent the interior of
    a unit declaration or reference (inside the brackets (excluded)).
    @pre: there is no syntax error in this part.
    """
    start_index = 0
    if tokens_inside_unit[0] == CASE_GEN_SYM:
        start_index = 1
    name = ""
    while (   start_index < len(tokens_inside_unit)
          and not is_special_sym(tokens_inside_unit[start_index])):
        name += tokens_inside_unit[start_index]
        start_index += 1
    return remove_escapement(name)

def find_words(tokens_inside_word_group):
    """
    Finds the words in the tokens that represent the interior of a word group.
    Returns the list of those words in sequence.
    @pre: there is no syntax error in this part.
    """
    words = []
    for token in tokens_inside_word_group:
        if token == CASE_GEN_SYM:
            continue
        if  token in (RAND_GEN_SYM, VARIATION_SYM, ARG_SYM):
            return words
        words.append(token)
    return words


def find_modifiers_decl(tokens_inside_decl):
    """
    Finds and create a representation of the modifiers from a list of tokens
    representing the inside of a unit declaration. Returns the representation.
    If the percentage of generation was present but couldn't be
    @pre: there is no syntax error in this part (except possibly for
          percentage of generation).
    """
    modifiers = mods.UnitDeclarationModifiersRepr()

    i = 0
    if tokens_inside_decl[0] == CASE_GEN_SYM:
        modifiers.casegen = True
        i += 1

    expecting_variation = False
    expecting_argument = False
    while i < len(tokens_inside_decl):
        if tokens_inside_decl[i] == VARIATION_SYM:
            modifiers.variation_name = ""
            expecting_variation = True
            expecting_argument = False
        elif tokens_inside_decl[i] == ARG_SYM:
            modifiers.argument_name = ""
            expecting_variation = False
            expecting_argument = True
        elif expecting_variation:
            modifiers.variation_name += tokens_inside_decl[i]
        elif expecting_argument:
            modifiers.argument_name += tokens_inside_decl[i]
        i += 1

    modifiers.variation_name = remove_escapement(modifiers.variation_name)
    modifiers.argument_name = remove_escapement(modifiers.argument_name)

    return modifiers

def find_modifiers_reference(tokens_inside_reference):
    """
    Finds and create a representation of the modifiers from a list of tokens
    representing the inside of a reference. Returns the representation.
    @pre: there is no syntax error in this part.
    """
    modifiers = mods.ReferenceModifiersRepr()

    i = 0
    if tokens_inside_reference[0] == CASE_GEN_SYM:
        modifiers.casegen = True
        i += 1

    expecting_randgen_name = False
    expecting_percentgen = False
    expecting_variation = False
    expecting_argument = False
    while i < len(tokens_inside_reference):
        if tokens_inside_reference[i] == RAND_GEN_SYM:
            modifiers.randgen_name = ""
            expecting_randgen_name = True
            expecting_percentgen = False
            expecting_variation = False
            expecting_argument = False
        elif tokens_inside_reference[i] == PERCENT_GEN_SYM:
            expecting_randgen_name = False
            expecting_percentgen = True
            expecting_variation = False
            expecting_argument = False
        elif tokens_inside_reference[i] == VARIATION_SYM:
            modifiers.variation_name = ""
            expecting_randgen_name = False
            expecting_percentgen = False
            expecting_variation = True
            expecting_argument = False
        elif tokens_inside_reference[i] == ARG_SYM:
            modifiers.argument_value = ""
            expecting_randgen_name = False
            expecting_percentgen = False
            expecting_variation = False
            expecting_argument = True
        elif expecting_randgen_name:
            modifiers.randgen_name += tokens_inside_reference[i]
        elif expecting_percentgen:
            modifiers.percentage_randgen = int(tokens_inside_reference[i])
            expecting_percentgen = False
        elif expecting_variation:
            modifiers.variation_name += tokens_inside_reference[i]
        elif expecting_argument:
            modifiers.argument_value += tokens_inside_reference[i]
        i += 1

    modifiers.randgen_name = remove_escapement(modifiers.randgen_name)
    modifiers.variation_name = remove_escapement(modifiers.variation_name)
    modifiers.argument_value = remove_escapement(modifiers.argument_value)

    return modifiers

def find_modifiers_word_group(tokens_inside_word_group):
    """
    Finds and create a representation of the modifiers from a list of tokens
    representing the inside of a word group. Returns the representation.
    @pre: there is no syntax error in this part.
    """
    modifiers = mods.WordGroupModifiersRepr()

    i = 0
    if tokens_inside_word_group[0] == CASE_GEN_SYM:
        modifiers.casegen = True
        i += 1

    expecting_randgen_name = False
    expecting_percentgen = False
    while i < len(tokens_inside_word_group):
        if tokens_inside_word_group[i] == RAND_GEN_SYM:
            modifiers.randgen_name = ""
            expecting_randgen_name = True
            expecting_percentgen = False
        elif tokens_inside_word_group[i] == PERCENT_GEN_SYM:
            expecting_percentgen = True
            expecting_randgen_name = False
        elif expecting_randgen_name:
            modifiers.randgen_name += tokens_inside_word_group[i]
        elif expecting_percentgen:
            modifiers.percentage_randgen = int(tokens_inside_word_group[i])
            expecting_percentgen = False
        i += 1

    modifiers.randgen_name = remove_escapement(modifiers.randgen_name)

    return modifiers

def find_modifiers_choice(tokens_inside_choice):
    """
    Finds and create a representation of the modifiers from a list of tokens
    representing the inside of a choice. Returns the representation.
    @pre: there is no syntax error in this part.
    """
    modifiers = mods.ChoiceModifiersRepr()

    if tokens_inside_choice[0] == CASE_GEN_SYM:
        modifiers.casegen = True
    if tokens_inside_choice[-1] == RAND_GEN_SYM:
        modifiers.randgen = True

    return modifiers


def find_nb_examples_asked(annotation_interior):
    """
    Returns the training and testing number of examples asked for an intent
    declaration as a tuple. Returns `None` if the numbers given are not numbers.
    @pre: there is no syntax error in the annotation.
    """
    if len(annotation_interior) == 0:
        return None
    nb_train = None
    nb_test = None

    if len(annotation_interior) == 1:
        nb_train = annotation_interior[0]
    else:
        expecting_train = False
        expecting_test = False
        for token in annotation_interior:
            if (    token not in (ANNOTATION_ASSIGNMENT_SYM, ANNOTATION_SEP)
                and not token.isspace()):
                if PATTERN_NB_TRAIN_EX_KEY.match(token):
                    expecting_train = True
                elif PATTERN_NB_TEST_EX_KEY.match(token):
                    expecting_test = True
                elif expecting_train:
                    nb_train = token
                    expecting_train = False
                elif expecting_test:
                    nb_test = token
                    expecting_test = False

    if nb_train is None and nb_test is None:
        return None

    if nb_train is not None:
        nb_train = nb_train.replace(ANNOTATION_IGNORED_SYM, "")
    if nb_test is not None:
        nb_test = nb_test.replace(ANNOTATION_IGNORED_SYM, "")

    try:
        nb_train = int(nb_train)
        if nb_test is None:
            nb_test = 0
        else:
            nb_test = int(nb_test)
    except ValueError:
        return None
    return (nb_train, nb_test)


def find_alt_slot_and_index(slot_rule_tokens):
    """
    Returns the index of the equal sign and the alt slot value as a 2-tuple,
    from the tokens representing a slot rule. Returns `None` if no alt slot
    value was found.
    @pre: there is no syntax error in this part.
    """
    try:
        index = slot_rule_tokens.index(ALT_SLOT_VALUE_NAME_SYM)
    except ValueError:
        return None
    if index+1 < len(slot_rule_tokens):
        i = index+1
        alt_slot_val = slot_rule_tokens[i]
        if alt_slot_val == ' ':
            alt_slot_val = ""
        i += 1
        while i < len(slot_rule_tokens):
            alt_slot_val += slot_rule_tokens[i]
            i += 1
        return (index, remove_escapement(alt_slot_val))
    return None


def next_choice_tokens(choice_interior_tokens):
    """
    Yields the next choice as a list of tokens in `choice_interior_tokens`.
    @pre: there is no syntax error in this part.
    """
    current_choice = []
    for (i, token) in enumerate(choice_interior_tokens):
        if token == CASE_GEN_SYM:
            continue
        elif token == RAND_GEN_SYM:
            if i == len(choice_interior_tokens)-1:  # Random generation symbol
                # NOTE: this should be changed if named randgen or percentgen
                #       is supported in the future.
                break
            else:  # Not a random generation symbol
                current_choice.append(token)
        elif token == CHOICE_SEP:
            yield current_choice
            current_choice = []
        else:
            current_choice.append(token)
    yield current_choice



def next_sub_rule_tokens(tokens):
    """
    Yields the next sub-rule from a rule
    represented as tokens (i.e. a list of str).
    @pre: `tokens` represents a valid rule.
    """
    current_sub_rule = []
    stop_with_char = None
    reading_sub_rule = False
    for token in tokens:
        if reading_sub_rule:
            if token == stop_with_char:
                current_sub_rule.append(token)
                yield current_sub_rule
                current_sub_rule = []
                stop_with_char = None
                reading_sub_rule = False
            else:
                current_sub_rule.append(token)
        else:  # Looking for the start of a sub-rule
            if is_start_unit_sym(token):  # Unit reference starting point
                current_sub_rule.append(token)
                reading_sub_rule = True
                stop_with_char = UNIT_CLOSE_SYM
            elif token == UNIT_OPEN_SYM:  # Word group starting point
                current_sub_rule.append(token)
                reading_sub_rule = True
                stop_with_char = UNIT_CLOSE_SYM
            elif token == CHOICE_OPEN_SYM:  # Word group starting point
                current_sub_rule.append(token)
                reading_sub_rule = True
                stop_with_char = CHOICE_CLOSE_SYM
            else:  # Word
                yield [token]


def is_sub_rule_word(sub_rule_tokens):
    """
    Returns `True` if the list of str `sub_rule_tokens` represents a word.
    @pre: considers `sub_rule_tokens` is never a single space.
    """
    return len(sub_rule_tokens) == 1
def is_sub_rule_word_group(sub_rule_tokens):
    """
    Returns `True` if the list of str `sub_rule_tokens`
    represents a word group.
    @pre: considers `sub_rule_tokens` to be a valid sub-rule.
    """
    return sub_rule_tokens[0] == UNIT_OPEN_SYM
def is_sub_rule_choice(sub_rule_tokens):
    """
    Returns `True` if the list of str `sub_rule_tokens`
    represents a choice.
    @pre: considers `sub_rule_tokens` to be a valid sub-rule.
    """
    return sub_rule_tokens[0] == CHOICE_OPEN_SYM
def is_sub_rule_alias_ref(sub_rule_tokens):
    """
    Returns `True` if the list of str `sub_rule_tokens`
    represents an alias reference.
    @pre: considers `sub_rule_tokens` to be a valid sub-rule.
    """
    return sub_rule_tokens[0] == ALIAS_SYM
def is_sub_rule_slot_ref(sub_rule_tokens):
    """
    Returns `True` if the list of str `sub_rule_tokens`
    represents a slot reference.
    @pre: considers `sub_rule_tokens` to be a valid sub-rule.
    """
    return sub_rule_tokens[0] == SLOT_SYM
def is_sub_rule_intent_ref(sub_rule_tokens):
    """
    Returns `True` if the list of str `sub_rule_tokens`
    represents an intent reference.
    @pre: considers `sub_rule_tokens` to be a valid sub-rule.
    """
    return sub_rule_tokens[0] == INTENT_SYM



from chatette.units import Example


class IntentExample(Example):

    def __init__(self, name, text=None, entities=None):# -> None:
        super(IntentExample, self).__init__(text, entities)
        self.name = name
    @classmethod
    def from_example(cls, name, ex):
        return cls(name, ex.text, ex.entities)

    # def __str__(self):
    #     return str(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.name+self.text+str(self.entities))



import io
import json
import os

from chatette.utils import cast_to_unicode
from chatette.units import ENTITY_MARKER
from ._base import Adapter


class JsonListAdapter(Adapter):
    def _get_file_extension(self):
        return "jsonl"

    def prepare_example(self, example):
        example.text = example.text.replace(ENTITY_MARKER, "")
        return json.dumps(cast_to_unicode(example.__dict__), ensure_ascii=False, sort_keys=True)

    def _write_batch(self, output_file_handle, batch):
        output_file_handle.writelines([
            self.prepare_example(example) + "\n"
            for example in batch.examples
        ])

    def write(self, output_directory, examples, synonyms):
        super(JsonListAdapter, self).write(output_directory, examples, synonyms)

        processed_synonyms = self.__synonym_format(synonyms)
        if processed_synonyms is not None:
            synonyms_file_path = os.path.join(output_directory, "synonyms.json")
            with io.open(synonyms_file_path, 'w') as output_file:
                output_file.write(json.dumps(cast_to_unicode(processed_synonyms),
                                             ensure_ascii=False,
                                             sort_keys=True, indent=2))


    @classmethod
    def __synonym_format(cls, synonyms):
        result = {key: values for (key, values)in synonyms.items()
                              if len(values) > 1 or values[0] != key}
        if not result:
            return None
        return result



"""
Module `chatette.cli.interactive_commands.rule_command`.
Contains the strategy class that represents the interactive mode command
`rule` which generates as many examples as asked that the provided rule
can generate.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.units.alias.definition import AliasDefinition
from chatette.units import ENTITY_MARKER


class RuleCommand(CommandStrategy):

    def execute(self, facade):
        """
        Implements the command `rule` which generates a certain number of
        examples according to a provided rule.
        """
        if len(self.command_tokens) < 2:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         'rule "<rule>" [<number-of-examples]')
            return

        rule_str = CommandStrategy.remove_quotes(self.command_tokens[1])
        nb_examples = None
        if len(self.command_tokens) >= 3:
            try:
                nb_examples = int(self.command_tokens[2])
            except ValueError:
                self.print_wrapper.error_log("The number of examples asked (" +
                                             self.command_tokens[2] + ") is " +
                                             "a valid integer.")

        rule_tokens = facade.parser.tokenizer.tokenize(rule_str)
        rule = facade.parser.tokens_to_sub_rules(rule_tokens)
        definition = AliasDefinition("INTERNAL", None, [rule])
        try:
            examples = definition.generate_nb_examples(nb_examples)
            self.print_wrapper.write("Generated examples:")
            for ex in examples:
                self.print_wrapper.write(ex.text.replace(ENTITY_MARKER, ""))
        except KeyError as e:
            self.print_wrapper.error_log("Upon generation: " + str(e))


    # Override abstract methods
    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self, facade):
        raise NotImplementedError()



"""
Module `chatette.cli.interactive_commands.exist_command`.
Contains the strategy class that represents the interactive mode command
`exist` which verifies whether a unit is declared or not.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class ExistCommand(CommandStrategy):
    usage_str = 'exist <unit-type> "<unit-name>"'

    def execute_on_unit(self, facade, unit_type, unit_name, variation_name=None):
        try:
            unit = facade.parser.get_definition(unit_name, unit_type)
            self.print_wrapper.write(unit.short_desc_str())
            if variation_name is not None:
                if variation_name in unit.variations:
                    self.print_wrapper.write("Variation '" + variation_name + \
                                             "' is defined for this " + \
                                             unit.type + ".")
                else:
                    self.print_wrapper.write("Variation '" + variation_name + \
                                             "' is not defined for this " + \
                                             unit.type + ".")
        except KeyError:
            self.print_wrapper.write(unit_type.name.capitalize() + " '" +
                                     unit_name + "' is not defined.")
        self.print_wrapper.write("")